
/*----------------------------------------------------------------------------+
 |                                                                            |
 |                    Enhanced HEYU Functionality                             |
 |           Copyright 2002,2003,2004,2005 Charles W. Sullivan                |
 |                      All Right Reserved                                    |
 |                                                                            |
 |                                                                            |
 | This software is licensed free of charge for non-commercial distribution   |
 | and for personal and internal business use only.  Inclusion of this        |
 | software or any part thereof in a commercial product is prohibited         |
 | without the prior written permission of the author.  You may copy, use,    |
 | and distribute this software subject to the following restrictions:        |
 |                                                                            |
 |  1)	You may not charge money for it.                                      |
 |  2)	You may not remove or alter this license, copyright notice, or the    |
 |      included disclaimers.                                                 |
 |  3)	You may not claim you wrote it.                                       |
 |  4)	If you make improvements (or other changes), you are requested        |
 |	to send them to the official Heyu maintainer, Charles W. Sullivan     |
 |      (cwsulliv01@heyu.org), so there's a focal point for distributing      |
 |      improved versions.                                                    |
 |                                                                            |
 | As used herein, HEYU is a trademark of Daniel B. Suthers, while X10,       | 
 | CM11A, ActiveHome, CM17A, and Firecracker are trademarks of X-10 (USA) Inc.|
 | The author is not affiliated with either entity.                           |
 |                                                                            |
 | Charles W. Sullivan                                                        |
 | Greensboro, North Carolina                                                 |
 | Email: cwsulliv01@heyu.org                                                 |
 |                                                                            |
 | Disclaimers:                                                               |
 | THERE IS NO ASSURANCE THAT THIS SOFTWARE IS FREE OF DEFECTS AND IT MUST    |
 | NOT BE USED IN ANY SITUATION WHERE THERE IS ANY CHANCE THAT ITS            |
 | PERFORMANCE OR FAILURE TO PERFORM AS EXPECTED COULD RESULT IN LOSS OF      |
 | LIFE, INJURY TO PERSONS OR PROPERTY, FINANCIAL LOSS, OR LEGAL LIABILITY.   |
 |                                                                            |
 | TO THE EXTENT ALLOWED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED "AS IS",|
 | WITH NO EXPRESS OR IMPLIED WARRANTY, INCLUDING, BUT NOT LIMITED TO, THE    |
 | IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.|
 |                                                                            |
 | IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL THE AUTHOR BE LIABLE    |
 | FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL   |
 | DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SOFTWARE EVEN IF   |
 | THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.            |
 |                                                                            |
 +----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#if defined(SYSV) || defined(FREEBSD)
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef NSLEEP
#include <sys/time.h>
#else
#include <time.h>
#endif
#include "x10.h"
#include "process.h"


extern int tty;
extern int line_no;
extern int sptty, i_am_child;
extern int verbose;
extern int xwrite(), exread(), check4poll();

extern CONFIG  config;

int special_func = 0;

char *typename[] = {"???", "Scene", "Usersyn"};

#if 0
/* Reference - Flags for parse_addr() function */
#define A_VALID   1
#define A_HCODE   2
#define A_BMAP    4
#define A_PLUS    8
#define A_MINUS  16
#define A_ALIAS  32
#define A_DUMMY  64
#define A_MULT  128
#endif
         

/* Arguments for the various commands */
static char *helparg[] = {
   "[H]",                /* 0 HC optional */
   "H",                  /* 1 HC only */
   "Hu",                 /* 2 HC | single unit */
   "HU",                 /* 3 HC | unit string */
   "HU <level>",         /* 4 Dim/Bright level*/
   "HU <level>",         /* 5 Preset dim level*/
   "<level>",            /* 6 Preset dim level only */
   "HU <level>",         /* 7 Extended Preset level*/
   "<T/F> HU <Data>",    /* 8 Arbitrary extended code command (hex)*/
   "<command ...>",      /* 9 Send HC|Funct only, no address */
   "xx xx xx ...",       /* 10 Arbitrary hex bytes */
   "H <\"text\">",       /* 11 Arbitrary quoted text message */
   "",                   /* 12 No arguments */
   "[<command>]",        /* 13 For help command */
   "N.NNN",              /* 14 Pause delay */
   "HU [HU [...]]",      /* 15 Multiple addresses */
   "HU <command>",       /* 16 For Turn command */
   "<query_command>",    /* 17 Special functions */
   "HU <linked>",        /* 18 Preset for macros */
   "HU <mode 0-3>",      /* 19 Ext status_ack mode */
   "<\"text\">",         /* 20 Message to log file */
   "NNN",                /* 21 Delay in minutes */
   "H[U] <count>",       /* 22 RF Dim/Bright */
   "xx xx <count>",      /* 23 Two hex bytes */
   "HU <count>",         /* 24 RF Dimbo */
   "H[u]",               /* 25 State commands */
#if 0   
   "H[U] <count>",       /* 26 RF Dim/Bright (fast mode) */
   "HU <count>",         /* 27 RF Dimbo (fast mode) */
#endif   
};

static char *options[] = {
    "Usage: heyu [options] <command>   (Run 'heyu help' for commands)\n",
    " [Options]",
    "    -v             Enable verbose mode",
    "    -c <pathname>  Specify full configuration file pathname",
    "    -s <pathname>  Specify full schedule file pathname",
    "    -0 ... -9      Config file is in subdirectory /0 ... /9",
    "                   of standard location, e.g., $HOME/.heyu/3/x10config",
};

/* Descriptions of "administrative" commands */
static char *helpadmin[][3] = {
    {"info","","Display CM11A registers, clock, and upload status"},
    {"date","","Return date in date(1) input format"},
    {"erase","","Zero CM11A EEPROM, erasing all events and macros"},
    {"help","[command]","This screen, or specific command help"},
    {"syn","","Display built-in synonyms for direct commands"},
    {"<label>","","Execute a scene or usersyn defined in the config file"},
    {"show","<option>","(Enter 'heyu show' to see options)"},
    {"monitor","","Monitor X10 activity (end with <BREAK>)"},
    {"engine","","Start or restart the Heyu state engine"},
    {"initstate","[H]","Zero entire state table or just housecode H"},
    {"initothers","","Zero cumulative address table"},
    {"reset","[H]","Reset interface to housecode H or default"},
    {"setclock","","Set CM11A clock to system clock (per schedule)"},
    {"readclock","","Display CM11A and system clocks (per schedule)"},
    {"newbattery","","Reset CM11A battery timer to zero"},
    {"purge","","Cancel pending delayed macros"},
    {"clear","","Clear CM11A registers"},
    {"list","","Display Lock and Spool directory names"},
    {"stop","","Stop the current relay and state engine daemons"},
    {"upload","[check|croncheck]","Upload schedule to CM11A or check schedule file"},
    {"upload","status|cronstatus","Display status of uploaded schedule"},
    {"catchup","","Emulate uploaded Timers from 0:00 to current time today"},
    {"macro","<label>","Emulate Macro in uploaded schedule"},
    {"utility","<option>","(Enter 'heyu utility' to see options)"},
    {"logmsg","<\"text\">","Display text message in log file and/or monitor"},
    {"wait","[timeout]","Wait until macro execution is completed"},
    {"cm10a_init","","Initialize CM10A interface. (For CM10A only!)"},
    {"version","","Display the Heyu version and exit"},
};

static char *helpstate[][3] = {
    {"onstate","H[u]","On-state bitmap of H or boolian Hu"},
    {"dimstate","H[u]","Dim-state bitmap of H or boolian Hu"},
    {"chgstate","H[u]","Changed-state bitmap of H or boolian Hu"},
    {"addrstate","H[u]","Addressed-state bitmap of H or boolian Hu"},
    {"statestr","H","State mini-bitmaps of all units as ASCII string"},
    {"dimlevel","Hu","Brighness level of module Hu as 0-100%"},
    {"rawlevel","Hu","Native level (0-210, 1-32, or 0-63) of module Hu"},
    {"heyu_state","Hu","Heyu script environment state bitmap (as integer)"},
    {"heyu_rawstate","Hu","Heyu script raw environment state bitmap (as integer)"},
    {"xtend_state","Hu","Xtend script environment state bitmap (as integer)"},
    {"fetchstate","","See man page heyu(1)"},
};

/* Descriptions of direct and uploaded macro commands */
static char *helpdirect[] = {
   /*  0 */ "Turn units ON",
   /*  1 */ "Turn units OFF",
   /*  2 */ "Turn All Lights ON",
   /*  3 */ "Turn All Lights OFF (**)",
   /*  4 */ "Turn All Units OFF",
   /*  5 */ "Dim units by <level> (1-22)",
   /*  6 */ "Dim units by <level> (1-22) after full bright",
   /*  7 */ "Brighten units by <level> (1-22)",
   /*  8 */ "Brighten units by <level> (1-22) after full bright",
   /*  9 */ "Request ON/OFF status (two-way modules)",
   /* 10 */ "Status Acknowledge ON",
   /* 11 */ "Status Acknowledge OFF",
   /* 12 */ "Preset units to <level> (1-32)",
   /* 13 */ "Preset to <level> (1-32) (function only)",
   /* 14 */ "Extended Preset <level> (0-63) (LM14A)",
   /* 15 */ "Extended All Units ON (LM14A)",
   /* 16 */ "Extended All Units OFF (LM14A)",
   /* 17 */ "Extended Status Request (LM14A)",
   /* 18 */ "Extended Status Acknowledge",
   /* 19 */ "Extended command - general",
   /* 20 */ "Extended command - general, await ack",
   /* 21 */ "Send arbitrary hex bytes as addresses",
   /* 22 */ "Send Housecode|Units addresses only",
   /* 23 */ "Send command function only",
   /* 24 */ "Hail other devices",
   /* 25 */ "Hail Acknowledge",
   /* 26 */ "Send quoted text (special protocol)",
   /* 27 */ "Data Transfer (function code 0xC)",
   /* 28 */ "Send All_Units_Off to All Housecodes",
   /* 29 */ "Extended Turn Units Full ON (LM14A)",
   /* 30 */ "Extended Turn Units Full OFF (LM14A)",
   /* 31 */ "This screen, or specific command help",
   /* 32 */ "Display synonyms for direct commands",
   /* 33 */ "Hail other devices, await ack",
   /* 34 */ "Legacy status command as above",
   /* 35 */ "Legacy preset command as above",
   /* 36 */ "Pause for N.NNN seconds",
   /* 37 */ "Turn Units 1-16 ON",
   /* 38 */ "Change state on|off|up|down [vv]",
   /* 39 */ "Legacy turn command as above",
   /* 40 */ "Request temperature (RCS compatible)",
   /* 41 */ "Send bytes directly to interface",
   /* 42 */ "Limited Preset for macros - see manpage",
   /* 43 */ "Extended config auto status mode (LM14A)",
   /* 44 */ "Request RCS compatible status",
   /* 45 */ "Delay for NNN minutes in macros",
   /* 46 */ "Pause until next tick of system clock",
   /* 47 */ "Transmit RF On",
   /* 48 */ "Transmit RF Off",
   /* 49 */ "Transmit RF Dims [after On]",
   /* 50 */ "Transmit RF Brights [after On]",
   /* 51 */ "Transmit RF All Lights On",
   /* 51 */ "Transmit RF All Lights Off",
   /* 53 */ "Transmit RF All Units Off",
   /* 54 */ "Transmit RF Arbitrary 2-byte hex code",
   /* 55 */ "Reset CM17A device",
   /* 56 */ "Transmit RF Dims after Off",
   
   /* 57 */ "Transmit RF Dims [after On] (#)",
   /* 58 */ "Transmit RF Brights [after On] (#)",
   /* 59 */ "Transmit RF Dims after Off (#)",
   /* 60 */ "Transmit RF Arbitrary 2-byte hex code (#)",
};

   
void display_helpnotes ( void )
{
   printf("\n  (*)  Not available for use in uploaded macros.\n");
   printf("  (**) Standard one-way plug-in Lamp Modules do NOT support this command.\n");

#ifdef HASCM17A   
   printf("  (#)  Fast CM17A command - see man x10cm17a(5) for configuration.\n");
   printf("\n Man pages: heyu(1), x10config(5), x10sched(5), x10scripts(5), x10cm17a(5).\n");
#else
   printf("\n Man pages: heyu(1), x10config(5), x10sched(5), x10scripts(5).\n");
#endif
   return;
}


/* Flags for x10command table below */
#define  F_ALL  0x00000001 /* "All" command; no unit code required on input */
#define  F_DED  0x00000002 /* Display extended code data byte in report */
#define  F_NOA  0x00000004 /* Suppress HC|Unit address byte transmission */
#define  F_BRI  0x00000008 /* Brighten before dimming */
#define  F_FAD  0x00000010 /* Force HC|Unit address byte(s) transmission */
#define  F_ARB  0x00000020 /* Use special format for arbitrary extended code */
#define  F_NMA  0x00000040 /* Unavailable as an uploadable macro */
#define  F_HID  0x00000080 /* Hidden command */
#define  F_STA  0x00000100 /* Status/xStatus Request (await status) */
#define  F_NUM  0x00000200 /* Arguments are all numbers */
#define  F_HLP  0x00000400 /* For help menu only - not a CM11A command */
#define  F_NMAI 0x00000800 /* Unavailable and ignored for uploaded macro */
#define  F_TRN  0x00001000 /* For Turn command */
#define  F_SPF  0x00002000 /* For special function processing */
#define  F_MAC  0x00004000 /* Special version for uploaded macros */
#define  F_FIR  0x00008000 /* CM17A (Firecracker) command */
#define  F_FFM  0x00010000 /* CM17A "fast" mode */

#define  F_RFS (F_FIR | F_NMA)
#define  F_RFF (F_FIR | F_NMA | F_FFM)

/*---------------------------------------------------------------------+
 | Master table of commands and synonyms.  The first of any grouping   |
 | having all fields identical except for the label is considered the  |
 | command; subsequent entries in the grouping are treated as built-in |
 | synonyms.                                                           |
 +---------------------------------------------------------------------*/
/* Names for arbitrary extended code functions */
#define XARBNAME   "xfunc"
#define XARBNAMEW  "xfuncw"
  
static struct cmd_list {
   char          *label;
   unsigned char code;
   int           length;   /* Length as a macro */
   int           minargs;
   int           maxargs;
   unsigned long flags;
   unsigned char subcode;  /* Type|command for extended code commands */
   unsigned char argtype;
   unsigned char help;
} x10command[] = {
   {"on",               2, 3, 1, 1, 0,                0,  3, 0 },   /* Simple ON command */
   {"turnon",           2, 3, 1, 1, 0,                0,  3, 0 },   /* Simple ON command */
   {"off",              3, 3, 1, 1, 0,                0,  3, 1 },   /* Simple OFF command */
   {"turnoff",          3, 3, 1, 1, 0,                0,  3, 1 },   /* Simple OFF command */
   {"bright",           5, 4, 2, 2, 0,                0,  4, 7 },   /* Brighten command */
   {"turnup",           5, 4, 2, 2, 0,                0,  4, 7 },   /* Brighten command */
   {"up",               5, 4, 2, 2, 0,                0,  4, 7 },   /* Brighten command */
   {"brighten",         5, 4, 2, 2, 0,                0,  4, 7 },   /* Brighten command */
   {"tbright",          5, 4, 2, 2, F_NMA | F_HID,    1,  4, 7 },   /* Brighten command - triggered */
   {"brightb",          5, 4, 2, 2, F_BRI,            0,  4, 8 },   /* Brighten command with brighten first */
   {"bbright",          5, 4, 2, 2, F_BRI,            0,  4, 8 },   /* Brighten command with brighten first */
   {"dim",              4, 4, 2, 2, 0,                0,  4, 5 },   /* Dim command without brighten first */
   {"turndown",         4, 4, 2, 2, 0,                0,  4, 5 },   /* Dim command without brighten first */
   {"down",             4, 4, 2, 2, 0,                0,  4, 5 },   /* Dim command without brighten first */
   {"dimb",             4, 4, 2, 2, F_BRI,            0,  4, 6 },   /* Dim command with brighten first */
   {"bdim",             4, 4, 2, 2, F_BRI,            0,  4, 6 },   /* Dim command with brighten first */
   {"tdim",             4, 4, 2, 2, F_NMA | F_HID,    1,  4, 5 },   /* Dim command - triggered */
   {"lightson",         1, 3, 1, 1, F_ALL,            0,  1, 2 },   /* All_Lights_On */
   {"lights_on",        1, 3, 1, 1, F_ALL,            0,  1, 2 },   /* All_Lights_On */
   {"all_lights_on",    1, 3, 1, 1, F_ALL,            0,  1, 2 },   /* All_Lights_On */
   {"alllightson",      1, 3, 1, 1, F_ALL,            0,  1, 2 },   /* All_Lights_On */
   {"lightsoff",        6, 3, 1, 1, F_ALL,            0,  1, 3 },   /* All_Lights_Off */
   {"lights_off",       6, 3, 1, 1, F_ALL,            0,  1, 3 },   /* All_Lights_Off */
   {"all_lights_off",   6, 3, 1, 1, F_ALL,            0,  1, 3 },   /* All_Lights_Off */
   {"alllightsoff",     6, 3, 1, 1, F_ALL,            0,  1, 3 },   /* All_Lights_Off */
   {"allon",           26,-1, 1, 1, F_ALL,            0,  1,37 },   /* Units 1-16 On  */
   {"all_on",          26,-1, 1, 1, F_ALL,            0,  1,37 },   /* Units 1-16 On  */
   {"all_units_on",    26,-1, 1, 1, F_ALL,            0,  1,37 },   /* Units 1-16 On  */
   {"allunitson",      26,-1, 1, 1, F_ALL,            0,  1,37 },   /* Units 1-16 On  */
   {"units_on",        26,-1, 1, 1, F_ALL,            0,  1,37 },   /* Units 1-16 On  */
   {"unitson",         26,-1, 1, 1, F_ALL,            0,  1,37 },   /* Units 1-16 On  */
   {"alloff",           0, 3, 1, 1, F_ALL,            0,  1, 4 },   /* All_Units_Off  */
   {"all_off",          0, 3, 1, 1, F_ALL,            0,  1, 4 },   /* All_Units_Off  */
   {"all_units_off",    0, 3, 1, 1, F_ALL,            0,  1, 4 },   /* All_Units_Off  */
   {"allunitsoff",      0, 3, 1, 1, F_ALL,            0,  1, 4 },   /* All_Units_Off  */
   {"units_off",        0, 3, 1, 1, F_ALL,            0,  1, 4 },   /* All_Units_Off  */
   {"unitsoff",         0, 3, 1, 1, F_ALL,            0,  1, 4 },   /* All_Units_Off  */
   {"turn",            27,-1, 1,-1, F_TRN,            0, 16,38 },   /* Turn command */
   {"_turn",           99,-1, 0, 0, F_HLP | F_NMA,    0, 16,39 },   /* Legacy turn command */
   
   {"preset",          10, 3, 2, 2, F_NMA,            0,  5,12 },   /* Old Preset (not for macro) */
   {"preset_dim",      10, 3, 2, 2, F_NMA,            0,  5,12 },   /* Old Preset (not for macro) */
   {"preset",          11, 3, 2, 2, F_NMA | F_HID,    0,  5,12 },   /* Old Preset (not for macro) */
   {"preset_dim",      11, 3, 2, 2, F_NMA | F_HID,    0,  5,12 },   /* Old Preset (not for macro) */
   {"mpreset",         10, 3, 2, 2, F_MAC,            0, 18,42 },   /* Limited old Preset for macros */
   {"macro_preset",    10, 3, 2, 2, F_MAC,            0, 18,42 },   /* Limited old Preset for macros */
   {"mpreset",         11, 3, 2, 2, F_MAC | F_HID,    0, 18,42 },   /* Limited old Preset for macros */
   {"macro_preset",    11, 3, 2, 2, F_MAC | F_HID,    0, 18,42 },   /* Limited old Preset for macros */
   {"_preset",         99,-1, 0, 0, F_HLP | F_NMA,    0,  5,35 },   /* Legacy preset command help */
   {"preset_level",    10, 3, 1, 1, F_NOA | F_NUM,    0,  6,13 },   /* Old Preset with no address byte */
   {"preset_level",    11, 3, 1, 1, F_NOA|F_NUM|F_HID, 0,  6,13 },   /* Old Preset with no address byte */
   {"status",          15, 3, 1, 1, F_STA,            0,  3, 9 },   /* Status Request */
   {"status_req",      15, 3, 1, 1, F_STA,            0,  3, 9 },   /* Status Request */
   {"_status",         99,-1, 0, 0, F_HLP | F_NMA,    0,  3,34 },   /* Legacy status command help */
   {"status_on",       13, 3, 1, 1, F_ALL,            0,  3,10 },   /* Status On */
   {"status_off",      14, 3, 1, 1, F_ALL,            0,  3,11 },   /* Status Off */
   {"hail",             8, 3, 0, 1, F_ALL,            0,  0,24 },   /* Hail Request */
   {"hail_req",         8, 3, 0, 1, F_ALL,            0,  0,24 },   /* Hail Request */
   {"hailw",            8, 3, 0, 1, F_ALL|F_STA|F_NMA, 0,  0,33 },   /* Hail Request, wait for ack */
   {"hail_ack",         9, 3, 0, 1, F_ALL,            0,  0,25 },   /* Hail Acknowledge */
   {"data_xfer",       12, 3, 1, 1, F_ALL,            0,  1,27 },   /* Extended data transfer (0xC) */

   {"xon",              7, 6, 1, 1, F_DED,         0xfe,  3,29 },   /* Extended Full ON (LM14A) */
   {"xoff",             7, 6, 1, 1, F_DED,         0xfd,  3,30 },   /* Extended Full OFF (LM14A) */
   {"xpreset",          7, 6, 2, 2, F_DED,         0x31,  7,14 },   /* Extended Preset Dim (LM14A) */
   {"ext_preset",       7, 6, 2, 2, F_DED,         0x31,  7,14 },   /* Extended Preset Dim (LM14A) */
   {"xdim",             7, 6, 2, 2, F_DED,         0x31,  7,14 },   /* Extended Preset Dim (LM14A) */
   {"xallon",           7, 6, 1, 1, F_ALL,         0x33,  1,15 },   /* Extended All Units On */
   {"xall_on",          7, 6, 1, 1, F_ALL,         0x33,  1,15 },   /* Extended All Units On */
   {"ext_all_on",       7, 6, 1, 1, F_ALL,         0x33,  1,15 },   /* Extended All Units On */
   {"xalloff",          7, 6, 1, 1, F_ALL,         0x34,  1,16 },   /* Extended All Units Off */
   {"xall_off",         7, 6, 1, 1, F_ALL,         0x34,  1,16 },   /* Extended All Units Off */
   {"ext_all_off",      7, 6, 1, 1, F_ALL,         0x34,  1,16 },   /* Extended All Units Off */
   {"xstatus",          7, 6, 1, 1, F_STA,         0x37,  3,17 },   /* Extended Status Request */
   {"ext_status",       7, 6, 1, 1, F_STA,         0x37,  3,17 },   /* Extended Status Request */
   {"xconfig",          7, 6, 2, 2, F_DED,         0x3b, 19,43 },   /* Extended Auto Status Mode */
   { XARBNAME,          7, 6, 3, 3, F_ARB,         0xff,  8,19 },   /* Extended code - arbitrary */
   { XARBNAMEW,         7, 6, 3, 3, F_ARB|F_STA|F_NMA, 0xff, 8,20 }, /* Extended Status Req - arbitrary */

   {"address",         20,-1, 1,-1, F_NMA,            0, 15,22 },   /* Address bytes only (*)*/
   {"function",        21,-1, 1,-1, F_NOA,            0,  9,23 },   /* HC|Func byte only */
   {"kill_all_hc",     24,-1, 0, 0, 0,                0, 12,28 },   /* Send All_Units_Off to all housecodes */
   {"pause",           25,-1, 1, 1, F_NMA | F_NUM,    0, 14,36 },   /* Pause (for use in scenes only)*/
   {"rcs_req",         28,-1, 1,-1, F_SPF | F_NMA,    0, 17,44 },   /* RCS-compatible general status query */
   {"temp_req",        28,-1, 1,-1, F_SPF | F_NMA,    0, 17,40 },   /* RCS-compatible temperature query */

   {"command2cm11a",   31,-1, 1,-1, F_NMA|F_NUM|F_HID, 0, 10,41 },  /* Send bytes directly to interface */
   {"bytes2cm11a",     29,-1, 1,-1, F_NMA|F_NUM|F_HID, 0, 10,41 },  /* Send bytes directly to interface */
   {"sendbytes",       22,-1, 1,-1, F_NMA|F_NUM|F_HID, 0, 10,21 },   /* Send arbitrary hex bytes (*)*/
   {"sendtext",        23,-1, 2, 2, F_NMA|F_ALL|F_HID, 0, 11,26 },   /* Send (quoted) text message (*)*/
   {"pausetick",       33,-1, 0, 0, F_NMA | F_HID,    0, 12,46 },   /* Pause until next tick of sys clock */
   {"delay",           32,-1, 1, 1, F_NUM | F_HID,    0, 21,45 },   /* Delay in minutes */

#ifdef HASCM17A   
   {"freset",          34,-1, 0, 0, F_RFS,            8, 12,55 },   /* CM17A Reset */
   {"fon",             34,-1, 1, 1, F_RFS,            2,  3,47 },   /* CM17A On RF command */
   {"foff",            34,-1, 1, 1, F_RFS,            3,  3,48 },   /* CM17A Off RF command */
   {"fbright",         34,-1, 2, 2, F_RFS,            5, 22,50 },   /* CM17A Bright RF command */
   {"fdim",            34,-1, 2, 2, F_RFS,            4, 22,49 },   /* CM17A Dim RF command */
   {"fdimbo",          34,-1, 2, 2, F_RFS,            9, 24,56 },   /* CM17A Dimb RF command */
   {"flightson",       34,-1, 1, 1, F_RFS | F_ALL,    1,  1,51 },   /* CM17A AllLightsOn RF command */
   {"flightsoff",      34,-1, 1, 1, F_RFS | F_ALL,    6,  1,52 },   /* CM17A AllLightsOff RF command */
   {"falloff",         34,-1, 1, 1, F_RFS | F_ALL,    0,  1,53 },   /* CM17A AllUnitsOff RF command */
   {"farb",            34,-1, 3, 3, F_RFS | F_NUM,    7, 23,54 },   /* CM17A Arbitrary RF command */
   {"ffbright",        34,-1, 2, 2, F_RFF,            5, 22,58 },   /* CM17A Fast Bright RF command */
   {"ffdim",           34,-1, 2, 2, F_RFF,            4, 22,57 },   /* CM17A Fast Dim RF command */
   {"ffdimbo",         34,-1, 2, 2, F_RFF,            9, 24,59 },   /* CM17A Dimb RF command */
   {"ffarb",           34,-1, 3, 3, F_RFF | F_NUM,    7, 23,60 },   /* CM17A Fast Arbit RF command */  
#endif
};
int nx10cmds = sizeof(x10command)/sizeof(struct cmd_list);

	
/* Function labels for monitor display */
/* (with fake add-on function "AllOn") */
char *funclabel[17] = {
   "AllOff", "LightsOn", "On", "Off", "Dim", "Bright", "LightsOff",
   "Extended", "Hail", "HailAck", "Preset", "Preset",
   "DataXfer", "StatusOn", "StatusOff", "StatusReq", "AllOn",
};

char *ext3funclabel[16] = {
   "", "xPreset", "", "xAllOn", "xAllOff", "", "", "xStatusReq", "xStatusAck",
   "", "", "xConfig", "", "", "", ""
};


char *rcs_temp ( int function, int predim, char hc, int unit )
{
   static char buffer[80];
   int         temp;
   
   if ( (function != 10 && function != 11) || unit < 11 ) {
      buffer[0] = '\0';
      return (char *)NULL;      
   }

   temp = -60 + (predim - 1) + 32 * (unit - 11);
   sprintf(buffer, " Temperature =%3d : Location %c\n",
      temp, toupper(hc));

   return buffer;
}
      
/*---------------------------------------------------------------------+
 | Return the length of a macro element corresponding to the           |
 | argument command code.                                              |
 +---------------------------------------------------------------------*/
int macro_element_length ( unsigned char cmdcode )
{
   int j, length = -1;

   
   for ( j = 0; j < nx10cmds; j++ ) {
      if ( x10command[j].code == cmdcode && x10command[j].length > 0 ) {
         length = x10command[j].length;
         break;
      }
   }
   return length;
}
 
/*---------------------------------------------------------------------+
 | Return 1 if argument command label is in the administrative command |
 | (help) table, otherwise return 0.                                   |
 +---------------------------------------------------------------------*/
int is_admin_cmd ( char *label )
{
   int j;

   for ( j = 0; j < (sizeof(helpadmin)/sizeof(char *)/3); j++ ) {
      if ( strcmp(label, helpadmin[j][0]) == 0 )
         return 1;
   }
   return 0;
}

/*---------------------------------------------------------------------+
 | Return 1 if argument command label is in the x10command[] table,    |
 | otherwise return 0.                                                 |
 +---------------------------------------------------------------------*/
int is_direct_cmd ( char *label )
{
   int j;

   for ( j = 0; j < nx10cmds; j++ ) {
      if ( strcmp(label, x10command[j].label) == 0 &&
           !(x10command[j].flags & F_HLP) )
         return 1;
   }
   return 0;
}
   
/*---------------------------------------------------------------------+
 | Return 1 if argument command label is a scene or is in the          |
 | x10command[] table, otherwise return 0.                             |
 +---------------------------------------------------------------------*/
int is_heyu_cmd ( char *label )
{

   if ( lookup_scene(config.scenep, label) >= 0 ) {
      return 1;
   }

   if ( is_direct_cmd(label) ) {
      return 1;
   }

   return 0;
}   


/*---------------------------------------------------------------------+
 | Sleep for argument milliseconds.                                    |
 +---------------------------------------------------------------------*/
void millisleep( long millisec )
{
   #ifdef NSLEEP
   struct timestruc_t tspec;

   tspec.tv_sec = millisec / 1000;
   tspec.tv_nsec = 1000000L * (millisec % 1000);

   while ( nsleep( &tspec, &tspec ) == -1 );
   #else
   struct timespec tspec;

   tspec.tv_sec = millisec / 1000;
   tspec.tv_nsec = 1000000L * (millisec % 1000);

   while ( nanosleep( &tspec, &tspec ) == -1 );
   #endif

   return;
}   

/*---------------------------------------------------------------------+
 | Sleep for argument microseconds.                                    |
 +---------------------------------------------------------------------*/
void microsleep( long microsec )
{
   #ifdef NSLEEP
   struct timestruc_t tspec;

   tspec.tv_sec = microsec / 1000000L;
   tspec.tv_nsec = 1000L * (microsec % 1000000L);

   while ( nsleep( &tspec, &tspec ) == -1 );
   #else
   struct timespec tspec;

   tspec.tv_sec = microsec / 1000000L;
   tspec.tv_nsec = 1000L * (microsec % 1000000L);

   while ( nanosleep( &tspec, &tspec ) == -1 );
   #endif

   return;
}   


/*---------------------------------------------------------------------+
 | Compare two commands.  Return 0 if identical except for label.      |
 +---------------------------------------------------------------------*/
static int cmp_commands( struct cmd_list *one, struct cmd_list *two )
{
   
   if ( !one || !two ||
        one->flags   != two->flags   ||
	one->code    != two->code    ||
	one->subcode != two->subcode ||
        one->length  != two->length  ||
        one->minargs != two->minargs ||
	one->maxargs != two->maxargs ||
	one->argtype != two->argtype ||
	one->help    != two->help        )
      return 1;
   return 0;
}


/*---------------------------------------------------------------------+
 | Parse the units string.  Pass back the X10 units bitmap through the |
 | argument list.  Return 0 if valid bitmap, otherwise write error msg |
 | to error handler and return 1.  Unit 0 is considered a valid unit   |
 | number but doesn't appear in the bitmap unless it's the only unit,  |
 | whereupon the bitmap will be 0.                                     |
 +---------------------------------------------------------------------*/
int parse_units ( char *str, unsigned int *bitmap )
{

   static int   umap[] = {6,14,2,10,1,9,5,13,7,15,3,11,0,8,4,12};
   char         buffer[256];
   char         errmsg[128];
   char         *tail, *sp;

   int          ulist[17];
   int          j, ustart, unit;

   clear_error_message();

   if ( strcmp(str, "*") == 0 ) {
      *bitmap = 0xffff;
      return 0;
   }

   for ( j = 0; j < 17; j++ )
      ulist[j] = 0;

   ustart = -1;
   tail = str;
   while ( *(get_token(buffer, &tail, "-,", 255)) ) {
      unit = (int)strtol(buffer, &sp, 10);
      if ( *sp ) {
         sprintf(errmsg, "Invalid character '%c' in units list", *sp);
         store_error_message(errmsg);
         return 1;
      }

      if ( unit < 0 || unit > 16 ) {
         sprintf(errmsg, "Unit number outside range 0-16");
         store_error_message(errmsg);
         return 1;
      }

      if ( *tail == ',' || *tail == '\0' ) {
         if ( ustart > -1 ) {
            if ( ustart <= unit ) {
               for ( j = ustart; j <= unit; j++ )
                  ulist[j] = 1;
               ustart = -1;
               continue;
            }
            else {
               for ( j = unit; j <= ustart; j++ )
                  ulist[j] = 1;
               ustart = -1;
               continue;
            }
         }
         else {
            ulist[unit] = 1;
            continue;
         }
      }
      else {
         ustart = unit;
      }
   }

   *bitmap = 0;
   for ( j = 0; j < 16; j++ )
      *bitmap |= (ulist[j+1]) << umap[j];

   return 0;
}

/*---------------------------------------------------------------------+
 | Return the number of units represented by a bitmap                  |
 +---------------------------------------------------------------------*/
int count_bitmap_units ( unsigned int bitmap )
{
   int j, mask, count;

   mask = 1;
   count = 0;
   for ( j = 0; j < 16; j++ ) {
      if ( bitmap & mask )
         count++;
      mask = mask << 1;
   }
   return count;
}
         
/*---------------------------------------------------------------------+
 | Parse the Houscode|Units token, which may be comprised of:          |
 |   Optional '+' or '-' prefix, plus:                                 |
 |   An ALIAS, or:                                                     |
 |     Housecode letter in range A-P, or '_' for placeholder.          |
 |     Heyu format units string, e.g., 2,3,5-8,10,13-15                |
 | Pass back the Housecode letter and X10 units bitmap through the     |
 | argument list.                                                      |
 | Return flags indicating the presense of each component:             |
 |   A_VALID    Valid address.                                         |
 |   A_HCODE    Housecode A-P has been returned.                       |
 |   A_BMAP     Non-zero bitmap has been returned                      |
 |   A_PLUS     Prefix '+' sign                                        |
 |   A_MINUS    Prefix '-' sign                                        |
 |   A_ALIAS    Housecode and bitmap came from an alias in config      |
 |   A_DUMMY    Replaceable address, e.g., $1, $3                      |
 |   A_MULT     Bitmap contains multiple units.                        |
 +---------------------------------------------------------------------*/
unsigned int parse_addr ( char *token, char *housecode,
                                                     unsigned int *bitmap )
{
   char          errmsg[128];
   unsigned int  flags = 0;
   char          *sp;
   char          hc;
   
   sp = token;

   if ( strchr(sp, '$') ) {
      /* Supply a fake value that will always pass scene verification */
      /* for a parameter containing a positional parameter.           */
      *housecode = 'A';
      *bitmap = 1;
      flags = (A_DUMMY | A_VALID | A_HCODE | A_BMAP | A_PLUS);
      return flags;
   }

   if ( *sp == '+' )  {
      flags |= A_PLUS;
      sp++;
   }
   else if ( *sp == '-' ) {
      flags |= A_MINUS;
      sp++;
   }

   if ( *sp == '\0' ) {
      *housecode = '_';
      *bitmap = 0;
      flags |= A_VALID;
      return flags;
   }


   if ( alias_lookup(sp, housecode, bitmap) >= 0 ) {
      flags |= (A_ALIAS | A_HCODE | A_VALID);
      if ( *bitmap )
         flags |= A_BMAP;
      if ( count_bitmap_units(*bitmap) > 1 )
         flags |= A_MULT;
      return flags;
   }

   hc = toupper(*sp);

   if ( hc == '_' )
      hc = config.housecode;

   if ( (hc != '_') && (hc < 'A' || hc > 'P') ) {
      *housecode = '_';
      *bitmap = 0;
      sprintf(errmsg, "Invalid housecode '%c' in HU '%s'", *sp, token);
      store_error_message(errmsg);
      flags &= ~(A_VALID | A_HCODE);
      return flags;
   }
   
   *housecode = hc;
   if ( hc != '_' )
      flags |= A_HCODE;
   sp++;

   if ( *sp == '\0' ) {
      *bitmap = 0;
      flags |= A_VALID;
      return flags;
   }
   else if ( parse_units(sp, bitmap) != 0 ) {
      /* Prefix the message written to error handler */
      sprintf(errmsg, "HU '%s': ", token);
      add_error_prefix(errmsg);
      return 0;
   }

   if ( *bitmap )
      flags |= A_BMAP;
   if ( count_bitmap_units(*bitmap) > 1 )
      flags |= A_MULT;

   flags |= A_VALID;

   return flags;
}

/*---------------------------------------------------------------+
 |  Return the 8-bit checksum of the bytes in argument buffer.   |
 +---------------------------------------------------------------*/     
unsigned char checksum( unsigned char *buffer, int length ) 
{
   int j;
   unsigned char sum = 0; 

   for ( j = 0; j < length; j++ )
      sum += buffer[j];

   return (sum & 0xff);
}


/*---------------------------------------------------------------+
 |  Return the 8-bit checksum of the bytes in argument buffer.   |
 |  The checksum starts with the first byte unless it is 0xFB    |
 |  and the length is 19 (for CM11A) or 43 (for CM10A), when     |
 |  the checksum starts with the second byte.                    |
 +---------------------------------------------------------------*/     
unsigned char checksum_special( unsigned char *buffer, int length ) 
{
   int j, jstart = 0;
   unsigned char sum = 0; 

   if ( buffer[0] == 0xfb &&
             length == ((config.device_type & DEV_CM10A) ? 43 : 19) ) 
      jstart = 1;
   else
      jstart = 0;
   
   for ( j = jstart; j < length; j++ )
      sum += buffer[j];

   return (sum & 0xff);
}

/*---------------------------------------------------------------+
 |  Send the contents of the buffer to the interface.            |
 |  Return 0 if successful, 1 otherwise.                         |
 +---------------------------------------------------------------*/  
int send_buffer ( unsigned char *buffer, int length,
                           int timeout, unsigned char triggermode )
{
   unsigned char chksum;
   unsigned char inbuff[16];
   int           nread;
   extern int    wait_external_trigger(void);    

   chksum = checksum(buffer, length);

   (void) xwrite(tty, (char *) buffer, length);

   /* get a check sum in reply */

   nread = exread(sptty, inbuff, 1, timeout);

   if( chksum == inbuff[0] ) {
      /* Checksum is OK; tell interface to transmit the code. */
      if ( verbose )
         fprintf(stderr, "Checksum confirmed\n");

      /* Defer sending until the appropriate half-cycle if so          */
      /* specified (assuming we have the external hardware installed). */
      if ( triggermode > 0 )  	 
         wait_external_trigger();

      (void) xwrite(tty, "\00" , 1);	/* WRMI (we really mean it) */
   }
   else  {
      if ( verbose )
         fprintf(stderr, "Checksum %02x NOT confirmed - should be %02x\n",
            inbuff[0], chksum);       
      return 1;
   }

   if ( (nread = exread(sptty, inbuff, 1, timeout)) >= 1 ) {
      if( inbuff[0] == 0x55 ) {
         /* interface is ready again */
         if ( verbose )
            fprintf(stderr, "Interface is ready.\n");
         return 0;
      }
      else if ( nread == 1 && chksum == 0x5a ) {
         if ( verbose )
            fprintf(stderr, "Checksum 0x5a workaround.\n");
         return 0;
      }
      else if ( nread == 1 && inbuff[0] == 0x5a ) {
#if 0	      
         check4poll(0,1);
#endif	 
         return 0;
      }
      else {
         fprintf(stderr,
          "Interface not ready - received %d bytes beginning %02x\n",
            nread, inbuff[0]);
         return 1;
      }
   }
   
   return 1;
}
 

/*---------------------------------------------------------------+
 |  Send an encoded housecode|bitmap to the interface.           |
 |  Try up to 3 times.  Return 0 if successful, otherwise 1.     |
 |  Argument housecode is a X10 encoded char A-P                 |
 |  Argument bitmap is encoded per X10                           |
 +---------------------------------------------------------------*/
int send_address ( unsigned char hcode, unsigned int bitmap,
                                      unsigned char triggermode )
{  
   unsigned int  mask;
   unsigned char buffer[4];
   int           j, k, result = 1;
   int           timeout = 10;
   
   if ( bitmap == 0 )
      return 0;

   /* Disable script launching for child process */
   if ( i_am_child ) 
      send_x10state_command(ST_LAUNCH, 0);
   else
      send_x10state_command(ST_LAUNCH, 1);

   hcode = hcode << 4;

   mask = 1;
   for ( j = 0; j < 16; j++ ) {
      if ( bitmap & mask ) {
         buffer[0] = 0x04;
         buffer[1] = hcode | j ;
	 
	 /* Kluge "fix" for checksum 5A problem. */
	 /* CM11A seems to disregard a bit in the Dim field. */
	 if ( checksum(buffer, 2) == 0x5A && config.fix_5a == YES )
	    buffer[0] = 0x0C;

         if ( verbose )
            fprintf(stderr, "Sending address bytes: %02x %02x\n",
               buffer[0], buffer[1]);

         for ( k = 0; k < 3; k++ ) {
            if ( (result = send_buffer(buffer, 2, timeout, triggermode)) == 0 )
               break;
            sleep(1);
         }

         /* Delay a minimum of 3 powerline cycles between commands */
         millisleep(60);  /* Use 60 msec for both 50 Hz and 60 Hz */

         if ( result != 0 ) {
            fprintf(stderr, "Unable to send address bytes\n");
            return 1;
         }
      }   
            
      mask = mask << 1;
   }

   return 0;
}

/*---------------------------------------------------------------+
 |  Send an X10 command buffer to the interface. Try up to 3     |
 |  times.  Return 0 if sucessful, otherwise 1.                  | 
 +---------------------------------------------------------------*/
int send_command ( unsigned char *buffer, int len, 
                unsigned int showstatus, unsigned char triggermode )
{
   int j, result = 1;
   int timeout = 10;

   if ( len == 0 )
      return 0;

   /* Disable script launching for child process */
   if ( i_am_child ) 
      send_x10state_command(ST_LAUNCH, 0);
   else
      send_x10state_command(ST_LAUNCH, 1);

   if ( verbose )
      printf("Script launch is %s\n", 
        (i_am_child) ? "disabled" : "enabled");
  

   if ( verbose ) {
      printf("Sending command bytes:");
      for ( j = 0; j < len; j++ )
         printf(" %02x", buffer[j]);
      printf("\n");
   }

   for ( j = 0; j < 3; j++ )  {
      if ( (result = send_buffer(buffer, len, timeout, triggermode)) == 0 ) 
         break;
      sleep(1);
   }
   if ( result != 0 ) {
      fprintf(stderr, "Unable to send command bytes\n");
      return 1;
   }

   if ( special_func > 0 ) {
      /* Command has called for special processing of */
      /* reply from module, e.g, decoding of preset   */
      /* reply from TempLinc into temperature         */
      for ( j = 0; j < config.spf_timeout; j++ ) {
         check4poll(0, 1);
      }
   }
   else if ( showstatus ) {
      /* Display status if expected */
      for ( j = 0; j < config.status_timeout; j++ ) {
         check4poll(1, 1);
      }
   }
   else {
         check4poll(0, 0);
   }

   /* Delay a minimum of 3 powerline cycles between commands */
   millisleep(60);  /* Use 60 msec for both 50 Hz and 60 Hz */

   if ( verbose )
      printf("Transmission OK\n");

   return 0;

}

/*---------------------------------------------------------------+
 | Ping the CM11A by sending the RI enable code.                 |
 +---------------------------------------------------------------*/
int c_ping ( int argc, char *argv[] )
{
   int  j, result = 1;
   unsigned char buf[2];

   buf[0] = 0xeb;

   for ( j = 0; j < 3; j++ )  {
      if ( (result = send_buffer(buf, 1, 1, 0)) == 0 ) 
         break;
      sleep(1);
   }

   if ( result == 0 )
      printf("Ping response OK\n");
   else
      printf("No response to ping.\n");

   check4poll(0, 0);

   return 0;
}
   
/*---------------------------------------------------------------+
 | Wait for CM11A to respond to "pings" or until timeout.        |
 | Return 1 if timeout or error, otherwise 0.                    | 
 +---------------------------------------------------------------*/
int c_busywait ( int argc, char *argv[] )
{
   int  j, timeout, nread;
   char *sp;
   char buf[2], inbuf[16];
   int xread();
   
   timeout = 30;
   
   if ( argc > 2 ) {
      timeout = (int)strtol(argv[2], &sp, 10);
      if ( !strchr(" \t\n", *sp) || timeout < 0 || timeout > 300 ) {
         fprintf(stderr, "Invalid %s timeout parameter.\n", argv[1]);
         return 1;
      }
   }

   buf[0] = 0xeb;

   send_x10state_command (ST_BUSYWAIT, 1);
   
   for ( j = 0; j < timeout; j++ ) {
      
      (void) write(tty, buf, 1);

      nread = xread(sptty, inbuf, 1, 1);

      if ( nread > 0 && inbuf[0] == buf[0] ) {
         return 0;
      }
      millisleep(60);
   }

   send_x10state_command (ST_BUSYWAIT, 0);
   check4poll(0, 0);
   return 1;
}  
   
/*---------------------------------------------------------------+
 | Display help for argument command, or all commands if empty   |
 | string.                                                       | 
 +---------------------------------------------------------------*/
#define COMPACT_FORMAT
void command_help ( char *command ) 
{
   int j;
   int szlbl, szarg, sztot;
   struct cmd_list *one = NULL, *two;
   char buffer[256];

   if ( *command == '\0' ) {
      /* Display options */
      for ( j = 0; j < (sizeof(options)/sizeof(char *)); j++ ) {
         printf("%s\n", options[j]);
      }
      printf("\n");

      /* Display all commands */

      /* Administrative commands */
      /* Get max string lengths for formatting output */
      szlbl = szarg = sztot = 0;
      for ( j = 0; j < (sizeof(helpadmin)/sizeof(char *)/3); j++ ) {
         szlbl = max(szlbl, strlen(helpadmin[j][0]));
         szarg = max(szarg, strlen(helpadmin[j][1]));
         sztot = max(sztot, (strlen(helpadmin[j][0]) + 
                                     strlen(helpadmin[j][1])));
      }
      printf(" [Administrative commands  (H = Housecode)]\n");
      for ( j = 0; j < (sizeof(helpadmin)/sizeof(char *)/3); j++ ) {
         #ifdef COMPACT_FORMAT
            /* Compact format */
            sprintf(buffer, "%s  %s", helpadmin[j][0], helpadmin[j][1]);    
            printf("  heyu  %-*s  %s\n", sztot + 2, buffer, helpadmin[j][2]);
         #else
            printf("  heyu  %-*s %-*s  %s\n", szlbl, helpadmin[j][0],
             szarg, helpadmin[j][1], helpadmin[j][2]);
         #endif  /* End ifdef */
      }
      printf("\n");

      /* State commands */
      /* Get max string lengths for formatting output */
      szlbl = szarg = sztot = 0;
      for ( j = 0; j < (sizeof(helpstate)/sizeof(char *)/3); j++ ) {
         szlbl = max(szlbl, strlen(helpstate[j][0]));
         szarg = max(szarg, strlen(helpstate[j][1]));
         sztot = max(sztot, (strlen(helpstate[j][0]) + 
                                     strlen(helpstate[j][1])));
      }
      printf(" [State commands  (H = Housecode, u = Single unit) - require heyu engine]\n");
      for ( j = 0; j < (sizeof(helpstate)/sizeof(char *)/3); j++ ) {
         #ifdef COMPACT_FORMAT
            /* Compact format */
            sprintf(buffer, "%s  %s", helpstate[j][0], helpstate[j][1]);    
            printf("  heyu  %-*s  %s\n", sztot + 2, buffer, helpstate[j][2]);
         #else
            printf("  heyu  %-*s %-*s  %s\n", szlbl, helpstate[j][0],
             szarg, helpstate[j][1], helpstate[j][2]);
         #endif  /* End ifdef */
      }
      printf("\n");

      /* Direct commands */
      /* Get max string lengths for formatting output */
      szlbl = szarg = sztot = 0;
      for ( j = 0; j < nx10cmds; j++ ) {
         if ( x10command[j].flags & F_HID )
            continue;
         sztot = max(sztot, (strlen(x10command[j].label) + 
                          strlen(helparg[x10command[j].argtype])) );
      }

      printf(" [Direct commands  (H = Housecode, U = Units list)]\n");        
      for ( j = 0; j < nx10cmds; j++ ) {
	 two = &x10command[j];
	 if ( x10command[j].flags & (F_HID | F_FIR) ) {
	    /* Command not shown in menu */
            continue;
	 }
         if ( cmp_commands(one, two) ) {
            #ifdef COMPACT_FORMAT
               sprintf(buffer, "%s  %s", x10command[j].label, helparg[x10command[j].argtype]);
               printf("  heyu  %-*s  %s%s\n", sztot + 2, buffer,
                   helpdirect[x10command[j].help],
                   (x10command[j].flags & (F_NMA | F_NMAI)) ? " (*)" : "");
            #else
               printf("  heyu  %-*s %-*s %s%s\n", szlbl, x10command[j].label,
                  szarg, helparg[x10command[j].argtype], 
	          helpdirect[x10command[j].help],
                  (x10command[j].flags & (F_NMA | F_NMAI)) ? " (*)" : "");
            #endif  /* End ifdef */ 
	    one = two;
	 }
      }
    
      #ifdef HASCM17A 
      if ( 1 ) { 
         printf("\n [CM17A \"Firecracker\" commands  (H = Housecode, U = Units list)]\n");        
         for ( j = 0; j < nx10cmds; j++ ) {
            two = &x10command[j];
	    if ( !(x10command[j].flags & F_FIR) ) {
               continue;
	    }
            if ( cmp_commands(one, two) ) {
               #ifdef COMPACT_FORMAT
                  sprintf(buffer, "%s  %s", x10command[j].label, helparg[x10command[j].argtype]);
                  printf("  heyu  %-*s  %s%s\n", sztot + 2, buffer,
                      helpdirect[x10command[j].help],
                      (x10command[j].flags & (F_NMA | F_NMAI)) ? " (*)" : "");
               #else
                  printf("  heyu  %-*s %-*s %s%s\n", szlbl, x10command[j].label,
                     szarg, helparg[x10command[j].argtype], 
	             helpdirect[x10command[j].help],
                     (x10command[j].flags & (F_NMA | F_NMAI)) ? " (*)" : "");
               #endif  /* End ifdef */ 
	       one = two;
	    }
         }
      }
      #endif
      
      display_helpnotes();

      return;
   }
   else {
      /* Display specific command */
      for ( j = 0; j < (sizeof(helpadmin)/sizeof(char *)/3); j++ ) {
         if ( strcmp(command, helpadmin[j][0]) == 0 ) 
            break;
      }
      if ( j < (sizeof(helpadmin)/sizeof(char *)/3) ) {
         printf("Usage: heyu  %s %s  %s\n", helpadmin[j][0], 
             helpadmin[j][1], helpadmin[j][2]);
         return;
      }

      for ( j = 0; j < (sizeof(helpstate)/sizeof(char *)/3); j++ ) {
         if ( strcmp(command, helpstate[j][0]) == 0 ) 
            break;
      }
      if ( j < (sizeof(helpstate)/sizeof(char *)/3) ) {
         printf("Usage: heyu  %s %s  %s\n", helpstate[j][0], 
             helpstate[j][1], helpstate[j][2]);
         return;
      }

      for ( j = 0; j < nx10cmds; j++ ) {
	 if ( strcmp(command, x10command[j].label) == 0 )
            break;
      }
      if ( j < nx10cmds ) {
         printf("Usage: heyu  %s %s   %s%s\n", x10command[j].label,
            helparg[x10command[j].argtype], helpdirect[x10command[j].help],
            (x10command[j].flags & (F_NMA | F_NMAI)) ? " (*)" : "");
      }
      else {
	 printf("Command '%s' not recognized.\n", command);
         return;
      }
      return;
   }
}

/*---------------------------------------------------------------+
 | Send a hail_ack (for use only by state engine)                |
 +---------------------------------------------------------------*/
void acknowledge_hail ( void )
{
   extern   char default_housecode;
   extern   int  i_am_child;
   int      save_child;
   unsigned char buf[4];
   unsigned char hcode;

   hcode = hc2code(default_housecode);

   buf[0] = 0x06;
   buf[1] = hcode << 4 | 9;

   /* Temporarily set i_am_child so that ack will be sent */
   /* as if from a script.                                */

   save_child = i_am_child;
   i_am_child = 1;
   send_command(buf, 2, 0, 0);
   i_am_child = save_child;

   return;
}

         
/*---------------------------------------------------------------+
 | Display a list of all commands and their synonyms.            |
 +---------------------------------------------------------------*/
void display_syn ( void ) 
{
   int j;
   int szlbl;
   char delim = ' ';
   struct cmd_list *one = NULL, *two;

   /* Get max label length for formatting output */
   szlbl = 0;
   for ( j = 0; j < nx10cmds; j++ ) {
      if ( x10command[j].flags & (F_HLP | F_HID) )
         continue;
      szlbl = max(szlbl, strlen(x10command[j].label));
   }

   printf(" %-*s   %s\n", szlbl, "Command","Built-In Synonyms");
   printf(" %-*s   %s",   szlbl, "-------","-----------------");

   for ( j = 0; j < nx10cmds; j++ ) {
      two = &x10command[j];
      if ( x10command[j].flags & (F_HLP | F_HID) ) {
         /* Don't display hidden or purely help commands */
         continue;
      }
      if ( cmp_commands(one, two) ) {
         printf("\n %-*s ", szlbl, x10command[j].label);
         one = two;
         delim = ' ';
      }
      else {
         printf("%c %s", delim, x10command[j].label);
         delim = ','; 
      }
   }
   printf("\n");
   return;
}

	       

void cmd_usage ( FILE *fd, int index )
{
   fprintf(fd, "Usage: heyu %s %s\n", 
        x10command[index].label, helparg[x10command[index].argtype]);
}

/*---------------------------------------------------------------+
 | Send to the CM11A or merely verify the syntax of the command  |
 | represented by the argc tokens in list argv[] according to    |
 | input argument 'mode':                                        |
 |   mode = CMD_VERIFY   Merely verify the syntax                |
 |   mode = CMD_RUN      Actually send the command               |
 | Return 0 if successful; 1 otherwise.                          |
 | (The CMD_VERIFY mode is intended for verifying the syntax     |
 | of scene commands in the x10config file.)                     | 
 +---------------------------------------------------------------*/

#define CMD_VERIFY  0
#define CMD_RUN     1

int direct_command( int argc, char *argv[], int mode )
{
	
#ifdef HASCM17A  /* Used only for CM17A commands */
    extern unsigned int rf_unit_code[16];
    extern unsigned int rf_func_code[7];
    extern int display_rf_xmit(unsigned char, unsigned int, int);
    extern int write_cm17a( unsigned int, int, unsigned char );
    extern int reset_cm17a( void );
    extern void rf_post_delay( void );
    extern void rf_farb_delay( void );
    int           precode;
    int           bursts, burstgroup, savebursts;
    int           byte1, byte2;
    unsigned int  rfword, rfonword;
    unsigned char rfhcode, rfmode = RF_SLOW;
#endif

    extern int    line_no;
    int           j, k, icmd, nargs, radix, subcode;
    unsigned int  mask, aflags, bitmap, sendmap;
    unsigned long flags;
    char          *label, *sp;
    char          hc;
    char          errmsg[128];
    int           dimval, dimcode, fullbright, delay, timeout;
    double        pause;
    unsigned char cmdbuf[16], cmdcode, hcode, switchv, switchcode, testbuf[128];
    int           minargs, maxargs; 
    unsigned int  xdata=0;
    static char   reorder[] = {6,14,2,10,1,9,5,13,7,15,3,11,0,8,4,12};
    unsigned char trigmode;

    timeout = 10;
    line_no = 0;
    flags = 0;
    special_func = 0;

    clear_error_message();

    /* Handle a few special cases */

    if ( strcmp(argv[0], "function") == 0 ) {
       /* Special case: "function" command */
       if ( argc < 2 ) {
          store_error_message("Command 'function': Too few parameters");
          return 1;
       }
       argv++;
       argc--;
       flags |= F_NOA;
    }
    else if ( strcmp(argv[0], "turn") == 0 ) {
       /* Special case: "turn" command */
       if ( argc < 3 ) {
          store_error_message("Command 'turn': Too few parameters");
          return 1;
       }
       argv++;
       argc--;
       flags |= F_TRN;
       /* Swap first two parameters */
       sp = argv[0];
       argv[0] = argv[1];
       argv[1] = sp;
    }
    else if ( strcmp(argv[0], "temp_req") == 0 || 
              strcmp(argv[0], "rcs_req")  == 0 ) {
       if ( argc < 2 ) {
	  sprintf(errmsg, "Command '%s': Too few parameters", argv[0]);
          store_error_message(errmsg);
          return 1;
       }
       special_func = (mode == CMD_VERIFY) ? 0 : 1;
       argv++;
       argc--;
       flags |= F_SPF;
    }
    nargs = argc - 1;
     
    /* Identify the command */
    for ( icmd = 0; icmd < nx10cmds; icmd++ ) {
       if ( !strcmp(argv[0], x10command[icmd].label) )
          break;
    }
    if ( icmd == nx10cmds ) {
       sprintf(errmsg, "Invalid command '%s'", argv[0]);
       store_error_message(errmsg);
       return 1;
    }

    label    = x10command[icmd].label;
    minargs  = x10command[icmd].minargs;
    maxargs  = x10command[icmd].maxargs;
    flags   |= x10command[icmd].flags;
    cmdcode  = x10command[icmd].code;
    subcode  = x10command[icmd].subcode;
    radix    = (flags & F_ARB) ?  16 : 10;

#ifndef HASCM17A 
    /* CM17A support not included */
    if ( flags & F_FIR ) {
       sprintf(errmsg,
	  "Command '%s': Heyu was compiled without CM17A support.", label);
       store_error_message(errmsg);
       return 1;
    }
#endif
    
    /* Help menu items are invalid in scenes */
    if ( flags & F_HLP && mode == CMD_VERIFY ) {
       sprintf(errmsg, "Command '%s' is not valid in scenes/usersyns", label);
       store_error_message(errmsg);
       return 1;
    }
    
    /* If this is just a help menu item, return */
    if ( flags & F_HLP )
       return 0;

    /* Only commands with command code <= 6  or code == 26 are valid with "turn" */
    if ( flags & F_TRN && cmdcode > 6 && cmdcode != 26 ) {
       sprintf(errmsg, 
         "Command '%s' cannot be used with the 'turn' command", label); 
       store_error_message(errmsg);
       return 1;
    }

    if ( nargs < minargs ) {
       sprintf(errmsg, "Command '%s': Too few parameters", label);
       store_error_message(errmsg);
       return 1;
    }
    if ( maxargs != -1 && nargs > maxargs ) {
       sprintf(errmsg, "Command '%s': Too many parameters", label);
       store_error_message(errmsg);
       return 1;
    }

    /* Handle a few special cases */

    if ( flags & F_ARB ) {
       /* If arbitrary extended code, get subcode and adjust argument count */
       if ( strchr(argv[1], '$') && mode == CMD_VERIFY) {  
          /* Fake a dummy parameter for scene verification */
          subcode = 0x30;
          sp = " ";
       }
       else { 
          subcode = (int)strtol(argv[1], &sp, radix);
       }

       if ( !strchr(" \t\n", *sp) || subcode < 0 || subcode > 0xff ) {
          sprintf(errmsg,
             "Command '%s': Extended function code '%s' outside range 00-0xff",
                 label, argv[1]);
          store_error_message(errmsg);
          return 1;
       }
       argv++;
       argc--;
       nargs--;
    }

    
    /* Default HC|Unit */
    hc = '_';
    bitmap = 0;
    aflags = A_VALID;

    if ( !(flags & F_NUM) && nargs > 0 ) {
       /* parse it as an alias or HC|Units token. */
       aflags = parse_addr(argv[1], &hc, &bitmap);
    }

    /* Make sure a positional  parameter didn't slip through */
    if ( aflags & A_DUMMY && mode == CMD_RUN ) {
       sprintf(errmsg,
          "Command '%s': Unresolved positional parameter '%s'",
             label, argv[1]);
       store_error_message(errmsg);
       return 1;
    }
       
    if ( !(aflags & A_VALID) ) {
       sprintf(errmsg, "Command '%s': ", label);
       add_error_prefix(errmsg);
       return 1;
    }

    if ( !(aflags & A_HCODE) && !(flags & F_NUM) && minargs > 0 )  {
       sprintf(errmsg, "Command '%s': Housecode required", label);
       store_error_message(errmsg);
       return 1;
    }

    switchv = cmdcode;

    switch ( switchv ) {

       case 21 : /* Function only - we should never get here */
       case 27 : /* Turn command - we should never get here */
       case 28 : /* Temp_Req - we should never get here */
       case 99 : /* Legacy commands - we should never get here */
          fprintf(stderr, "Internal error - direct_command() - case %d\n", switchv);
          return 1;

       case  8 :  /* Hail */
       case  9 :  /* Hail Ack */
          if ( !(aflags & A_HCODE) ) {
             hc = config.housecode;
          }
       case  0 :  /* All Units Off  */
       case  1 :  /* All Lights On  */
       case  6 :  /* All Lights Off */
       case 12 :  /* Data Transfer */
          if ( bitmap != 0 && !(aflags & A_PLUS) && !(aflags & A_MINUS) && config.force_addr == NO ) {
             sprintf(errmsg, "Unitcode ignored for '%s' command.", label);
             store_error_message(errmsg);
             bitmap = 0;
          }
          if ( bitmap == 0 && (aflags & A_PLUS || config.force_addr == YES) )
             bitmap = 1;

       case  2 :  /* On */
       case  3 :  /* Off */
       case 13 :  /* Status On */
       case 14 :  /* Status Off */
          if ( aflags & A_MINUS || flags & F_NOA )
             bitmap = 0;
          else if ( bitmap == 0 && !(flags & F_ALL) ) {
             sprintf(errmsg, "command '%s': No Unitcode specified", label);
             store_error_message(errmsg);
             return 1;
          }

          if ( mode == CMD_VERIFY )
             break;

          hcode = hc2code(hc);

          cmdbuf[0] = 0x06;
          cmdbuf[1] = (hcode << 4) | cmdcode;

                 
          if ( send_address(hcode, bitmap, 0) != 0 ||
               send_command(cmdbuf, 2, (flags & F_STA), 0) != 0 )
             return 1; 
          
          break;

       case  4 :  /* Dim */
       case  5 :  /* Bright */
          if ( strchr(argv[2], '$') && mode == CMD_VERIFY ) {
             /* Fake a dim value for a dummy parameter */
             dimval = 10;
             sp = " ";
          }
          else {
             dimval = (int)strtol(argv[2], &sp, 10);
          }

          if ( !strchr(" \t\n", *sp) ) {
	     sprintf(errmsg,
		"Command '%s': Invalid Dim/Bright parameter '%s'", label, argv[2]);
             store_error_message(errmsg);
	     return 1;
	  }
	  else if ( config.restrict_dims == YES && (dimval < 1 || dimval > 22) ) {
             sprintf(errmsg, 
                "Command '%s': Dim/Bright value outside range 1-22", label);
             store_error_message(errmsg);
	     return 1;
	  }
	  else if ( config.restrict_dims == NO && (dimval < 0 || dimval > 31) ) {
             sprintf(errmsg, 
                "Command '%s': Dim/Bright value outside range 0-31", label);
             store_error_message(errmsg);
             return 1;
          }

          if ( aflags & A_MINUS || flags & F_NOA )
             bitmap = 0;
          else if ( bitmap == 0 ) {
             sprintf(errmsg, "Command '%s': no Unitcode specified", label);
             store_error_message(errmsg);
             return 1;
          }

          if ( mode == CMD_VERIFY )
             break;

          hcode = hc2code(hc);

	  fullbright = 22;

          trigmode = subcode;
	  
	  /* Kluge for housecode D to circumvent 0x5A checksum problem */
	  /* when sending dim Dx 22. Set dim value to 23 instead.      */ 
	  if ( hcode == hc2code('D') && config.fix_5a == YES ) {
	     fullbright = 23;
	     dimval = (dimval == 22) ? 23 : dimval;
	  }
	  
          /* If "brighten before dimming" flag is set, emulate a CM11A macro */
          /* by first brightening to full brightness.                        */
          if ( flags & F_BRI ) {
             cmdbuf[0] = (fullbright << 3) | 0x06;
             cmdbuf[1] = (hcode << 4) | 5;

             cmdbuf[2] = (dimval << 3) | 0x06;
             cmdbuf[3] = (hcode << 4) | cmdcode;
             
             if ( send_address(hcode, bitmap, trigmode) != 0 ||
                  send_command(cmdbuf, 2, (flags & F_STA), trigmode) != 0 ||
                  send_command(cmdbuf + 2, 2, (flags & F_STA), trigmode) != 0 )
                return 1;
          }
          else {
             cmdbuf[0] = (dimval << 3) | 0x06;
             cmdbuf[1] = (hcode << 4) | cmdcode;

             if ( send_address(hcode, bitmap, trigmode) != 0 ||
                  send_command(cmdbuf, 2, (flags & F_STA), trigmode) != 0 )
                return 1;
          }

          break;

       case 10 :  /* Old style preset 1 or preset_level 1 */
       case 11 :  /* Old style preset 2 or preset_level 2 */
          if ( aflags & A_MINUS || flags & F_NOA )
             bitmap = 0;
          else if ( bitmap == 0 ) {
             sprintf(errmsg, "Command '%s': no Unitcode specified", label);
             store_error_message(errmsg); 
             return 1;
          }

          /* (nargs = 2 for preset; nargs = 1 for preset_level) */

          if ( strchr(argv[nargs], '$') && mode == CMD_VERIFY ) {
             /* Fake a dim dummy parameter for scene verification */
             dimval = 10;
             sp = " ";
          }
          else
             dimval = (int)strtol(argv[nargs], &sp, 10);

          if ( !strchr(" \t\n", *sp) || dimval < 1 || dimval > 32 ) {
             sprintf(errmsg, "Command '%s': dim level outside range 1-32", label);
             store_error_message(errmsg);
             return 1;
          }

          if ( mode == CMD_VERIFY )
             break;

          if ( (dimcode = dimval) > 16 ) {
             cmdcode = 11;
             dimcode -= 16;
          }
          dimcode = rev_low_nybble(dimcode - 1);

          hcode = hc2code(hc);

          if ( flags & F_MAC && dimcode != hcode ) {
             sprintf(errmsg, "Command '%s': Preset level %d not supported for housecode %c",
                label, dimval, hc);
             store_error_message(errmsg);
             return 1;
          }

          cmdbuf[0] = 0x06;
          cmdbuf[1] = (dimcode << 4) | cmdcode;

          if ( send_address(hcode, bitmap, 0) != 0 ||
               send_command(cmdbuf, 2, (flags & F_STA), 0) != 0 )
             return 1;

          break;

       case 15 :  /* Status Req */
          sendmap = bitmap;
          if ( aflags & A_MINUS || flags & F_NOA ) {
             bitmap = 1;
             sendmap = 0;
          }
          else if ( bitmap == 0 ) {
             sprintf(errmsg, "Command '%s': no Unitcode specified", label);
             store_error_message(errmsg); 
             return 1;
          }

          if ( mode == CMD_VERIFY )
             break;

          hcode = hc2code(hc);
             
          cmdbuf[0] = 0x06;
          cmdbuf[1] = (hcode << 4) | cmdcode;

          if ( flags & F_SPF )
             flags &= ~F_STA;

          /* Send each address/command pair individually to allow */
          /* associating a response with a particular unit code   */
          mask = 1;
          for ( k = 0; k < 16; k++ )  {
             if ( bitmap & mask ) {
                if ( send_address(hcode, sendmap & mask, 0) != 0 ||
                     send_command(cmdbuf, 2, (flags & F_STA), 0) != 0 )
                   return 1;
             }
             mask = mask << 1;
          }         
          break;

       case 20 : /* Send address bytes only */
          /* Start over again with the address tokens */
          for ( k = 1; k < argc; k++ ) {
             aflags = parse_addr(argv[k], &hc, &bitmap);
             if ( !(aflags & A_VALID) ) {
                sprintf(errmsg, "Command '%s': ", label);
                add_error_prefix(errmsg);
                return 1;
             }
             else if ( !(aflags & A_HCODE) || bitmap == 0 ) {
                sprintf(errmsg, 
                   "Command '%s': invalid HU or Alias '%s'", label, argv[k]);
                store_error_message(errmsg); 
                return 1;
             }

             if ( mode == CMD_VERIFY )
                continue;

             hcode = hc2code(hc);

             if ( send_address(hcode, bitmap, 0) != 0 )
                return 1;

             /* A delay between addresses may be needed for */
             /* programming some SwitchLinc devices.  (The  */
             /* TempLinc programs OK without it.)           */
             millisleep(60);
          }
          break;

       case 29 : /* Send bytes directly to cm11a (highly experimental) */
       case 31 : /* Send bytes directly to cm11a for transmission (highly experimental) */
          /* First check that each arg is a byte */
          for ( k = 1; k < argc; k++ ) {
             if ( strchr(argv[k], '$') && mode == CMD_VERIFY ) {
                /* Fake a dummy value for scene verification */
                xdata = 10;
                sp = " ";
             }
             else {
                xdata = (int)strtol(argv[k], &sp, 16);
             }
             if ( !strchr(" \t\n", *sp) || xdata < 0 || xdata > 0xff ) {
                sprintf(errmsg, 
                   "Command '%s': Data '%s' outside range 0-FF", label, argv[k]);
                store_error_message(errmsg);
                return 1;
             }
          }

          if ( mode == CMD_VERIFY )
             break;

          for ( k = 1; k < argc; k++ ) {
             xdata = (int)strtol(argv[k], NULL, 16);
             testbuf[k-1] = xdata;
          }

	  if ( switchv == 29 )
             (void) xwrite(tty, (char *) testbuf, argc - 1);
          else
	     send_command(testbuf, argc - 1, 0, 0);
          break;


       case 22 : /* Send arbitrary series of hex bytes as addresses */
          /* First check that each byte is in range 00-FF */
          for ( k = 1; k < argc; k++ ) {
             if ( strchr(argv[k], '$') && mode == CMD_VERIFY ) {
                /* Fake a dummy value for scene verification */
                xdata = 10;
                sp = " ";
             }
             else {
                xdata = (int)strtol(argv[k], &sp, 16);
             }
             if ( !strchr(" \t\n", *sp) || xdata < 0 || xdata > 0xff ) {
                sprintf(errmsg, 
                   "Command '%s': Data '%s' outside range 0-FF", label, argv[k]);
                store_error_message(errmsg);
                return 1;
             }
          }

          if ( mode == CMD_VERIFY )
             break;

          for ( k = 1; k < argc; k++ ) {
             xdata = (int)strtol(argv[k], NULL, 16);
             if ( send_address(xdata >> 4, (1 << (xdata & 0x0f)), 0) != 0 )
                return 1;
          }
          break;

       case 23 : /* Send (quoted) text message */
          /* Format:                                                          */
          /* Send each character as two address bytes - high nybble first     */
          if ( !(aflags & A_DUMMY) && bitmap != 0 ) {
             sprintf(errmsg, "Command '%s': Unitcode ignored", label);
             store_error_message(errmsg);
             bitmap = 0;
          } 
          sp = argv[2];

          if ( mode == CMD_VERIFY )
             break;

          hcode = hc2code(hc);

          printf("Sending %d characters as Hcode|hi Hcode|low bytes:\n", 
                strlen(sp));

          while ( *sp != '\0' ) {
             if ( send_address(hcode, (1 << (*sp >> 4)), 0) != 0 ||
                  send_address(hcode, (1 << (*sp & 0x0f)), 0) != 0 )
                return 1;
             printf("%c", *sp);
             fflush(stdout);
             sp++;
          }
          printf("\n");
          break;

       case 24 :  /* Send All_Units_Off to all Housecodes */
          if ( mode == CMD_VERIFY )
             break;

          bitmap = (config.force_addr == YES) ?  1 : 0;
          cmdbuf[0] = 0x06;
	  for ( j = 0; j < 16; j++ ) {
	     hcode = reorder[j];
	     cmdbuf[1] = (hcode << 4);
	     if ( send_address(hcode, bitmap, 0) != 0 ||
                  send_command(cmdbuf, 2, 0, 0) != 0 )
                return 1;
             millisleep(60);	     
	  }
	  break;

       case 25 : /* Pause (for use in scenes) */
          if ( strchr(argv[1], '$') && mode == CMD_VERIFY ) {
             /* Fake a dummy parameter for scene verification */
             pause = 1.0;
             sp = " ";
          }
          else {
             pause = strtod(argv[1], &sp);
          }
          if ( !strchr(" \t\n", *sp) || pause < 0.0 ) {
             sprintf(errmsg, "Command '%s': invalid Pause value '%s'", label, argv[1]);
             store_error_message(errmsg);
             return 1;
          }

          if ( mode == CMD_VERIFY )
             break;

          if ( verbose )
             printf("pause %.3f seconds\n", pause);
          millisleep((int)(1000. * pause));

          break;
      
       case 32 : /* Delay in minutes */
          if ( strchr(argv[1], '$') && mode == CMD_VERIFY ) {
             /* Fake a dummy parameter for scene verification */
             delay = 1;
             sp = " ";
          }
          else {
             delay = strtol(argv[1], &sp, 10);
          }
          if ( !strchr(" \t\n", *sp) || delay < 0 || delay > 240 ) {
             sprintf(errmsg, "Command '%s': invalid Delay value '%s'", label, argv[1]);
             store_error_message(errmsg);
             return 1;
          }

          if ( mode == CMD_VERIFY )
             break;

          if ( verbose )
             printf("delay %d minutes\n", delay);
          millisleep((int)(60000. * delay));

          break;

       case 33 :  /* Wait for next tick of system clock */
	  if ( mode == CMD_VERIFY )
	     break;

	  wait_next_tick();

	  break;
      
       case 26 :  /* All Units On */
          if ( bitmap != 0 && !(aflags & A_MINUS) ) {
             sprintf(errmsg, "Unitcode ignored for '%s' command.", label);
             store_error_message(errmsg);
          }
          bitmap = 0xFFFF;

          if ( mode == CMD_VERIFY )
             break;

          hcode = hc2code(hc);

          cmdbuf[0] = 0x06;
          cmdbuf[1] = (hcode << 4) | 2 ;
                 
          if ( send_address(hcode, bitmap, 0) != 0 ||
               send_command(cmdbuf, 2, (flags & F_STA), 0) != 0 )
             return 1; 
          
          break;

       case 30 :  /* Show X10 state */
          if ( mode == CMD_VERIFY )
             break;

          hcode = hc2code(hc);

          cmdbuf[0] = 0x7b;
          cmdbuf[1] = 0;
          cmdbuf[2] = hcode << 4;

          xwrite(tty, cmdbuf, 3);
          millisleep(100);

          break;
	  
       case  7 :  /* Extended Code commands */
          sendmap = (aflags & A_PLUS) ? bitmap : 0 ;
          if ( sendmap == 0 && config.force_addr == YES ) 
             sendmap = 0xFFFF;

          hcode = hc2code(hc);

          if ( flags & F_ARB ) 
             switchcode = 0xff;
          else
             switchcode = subcode;

          switch ( switchcode ) {
             case 0x31 :  /* Extended Preset Dim */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': no Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }

                /* Get dim level */
                if ( strchr(argv[2], '$') && mode == CMD_VERIFY ) {
                   /* Fake a dummy parameter for scene verification */
                   xdata = 10;
                   sp = " ";
                }
                else {
                   xdata = (int)strtol(argv[2], &sp, 10);
                }
  
                if ( !strchr(" \t\n", *sp) || xdata < 0 || xdata > 63 ) {
                   sprintf(errmsg,
                     "Command '%s': Extended Preset level outside range 0-63", label);
                   store_error_message(errmsg);
                   return 1;
                }
                break;

             case 0x3b : /* Set module auto status_ack mode */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': no Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }

                /* Get mode (0-3) */
                if ( strchr(argv[2], '$') && mode == CMD_VERIFY ) {
                   /* Fake a dummy parameter for scene verification */
                   xdata = 3;
                   sp = " ";
                }
                else {
                   xdata = (int)strtol(argv[2], &sp, 10);
                }
  
                if ( !strchr(" \t\n", *sp) || xdata < 0 || xdata > 3 ) {
                   sprintf(errmsg,
                     "Command '%s': Extended Config Mode outside range 0-3", label);
                   store_error_message(errmsg);
                   return 1;
                }
                break;

             case 0xfe :  /* Extended ON ( = Extended Preset Dim level 63) */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': no Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }
                subcode = 0x31;
                xdata = 63;
                break;                

             case 0xfd :  /* Extended OFF ( = Extended Preset Dim level 0) */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "command '%s': no Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }
                subcode = 0x31;
                xdata = 0;
                break;                

             case 0x33 :  /* Extended All On */
             case 0x34 :  /* Extended All Off */

                if ( bitmap != 0 && !(aflags & A_PLUS) && !(aflags & A_MINUS) ) {
                   sprintf(errmsg, "Command '%s': Unitcode ignored", label);
                   store_error_message(errmsg);
                   bitmap = 1;
                }
                else if ( bitmap == 0 )
                   bitmap = 1;

                if ( aflags & A_PLUS )
                   sendmap = bitmap;

             case 0x37 :  /* Extended Status Req */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': no Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }
                xdata = 0;
                break;

             case 0xff :  /* Arbitrary Extended Function */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': no Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }
                if ( strchr(argv[2], '$') && mode == CMD_VERIFY ) {
                   /* Fake a dummy parameter for scene verification */
                   xdata = 10;
                   sp = " ";
                }
                else {
                   xdata = (int)strtol(argv[2], &sp, 16);
                }
                if ( !strchr(" \t\n", *sp) || xdata < 0 || xdata > 0xff ) {
                   sprintf(errmsg,
                      "Command '%s': Data '%s' outside range 0-FF", label, argv[2]);
                   store_error_message(errmsg); 
                   return 1;
                }
                break ;

             default :
                sprintf(errmsg, "Command '%s': Not implemented", label);
                store_error_message(errmsg);
                return 1;                
          }

          if ( mode == CMD_VERIFY )
             break;

          /* Header for extended code */
          cmdbuf[0] = 0x07;

          /* House code and extended code function */
          cmdbuf[1] = (hcode << 4) | 0x07 ;

          /* Unit code */
          /* cmdbuf[2] = unit code is filled in below */

          /* "data" value */
          cmdbuf[3] = (unsigned char)(xdata & 0xff);

          /* Extended Type/command */
          cmdbuf[4] = (unsigned char)(subcode & 0xff);


          mask = 1;
          for ( j = 0; j < 16; j++ ) {
             if ( bitmap & mask ) {
                cmdbuf[2] = j;

		/* Kluge "fix" for checksum 5A problem.    */
		/* CM11A seems to disregard the dim field  */
		/* in the header byte.                     */
		if ( checksum(cmdbuf, 5) == 0x5A && config.fix_5a == YES )
		   cmdbuf[0] = 0x0F;
		
                if ( send_address(hcode, sendmap & mask, 0) != 0 ||
                     send_command(cmdbuf, 5, flags & F_STA, 0) != 0 )
                   return 1;
             }
             mask = mask << 1;
          }
             
          break;

#ifdef HASCM17A	  
       case 34 : /* CM17A "Firecracker" RF commands */
	     
	  if ( aflags & A_PLUS && !(aflags & A_DUMMY) ) {
	     sprintf(errmsg,
	       "Command '%s': '+' prefix invalid for RF commands.", label);
	     store_error_message(errmsg);
	     return 1;
	  }
	  if ( flags & F_NOA ) {
             sprintf(errmsg, "'function' is invalid for RF commands.");
             store_error_message(errmsg);
             return 1;	     
	  }
	  if ( flags & F_FFM )
	     rfmode = RF_FAST;
		     
	  switch ( subcode ) {
	     case 0 :  /* RF All Units Off */
	     case 1 :  /* RF All Lights On */
	     case 6 :  /* RF All Lights Off */
		if ( bitmap != 0 && !(aflags & A_MINUS) ) {
                   sprintf(errmsg, "Unitcode ignored for '%s' command.", label);
                   store_error_message(errmsg);
		}

		if ( mode == CMD_VERIFY )
		   return 0;
		
                bursts = config.rf_bursts[subcode];
		
		rfhcode = rev_low_nybble(hc2code(hc));
		
		rfword = (rfhcode << 12) | rf_func_code[subcode];
                display_rf_xmit(subcode, rfword, bursts);
		if ( write_cm17a(rfword, bursts, rfmode) != 0 ) {
                   sprintf(errmsg,
		      "Command '%s': Unable to write to CM17A device.", label);
                   store_error_message(errmsg);
                   return 1;
                }
		rf_post_delay();

		break;

	     case 2 :  /* RF On */
	     case 3 :  /* RF Off */
		if ( aflags & A_MINUS ) {
		   sprintf(errmsg,
		     "Command '%s': '-' prefix invalid for this command.\n", label);
		   store_error_message(errmsg);
		   return 1;
                }
		
		if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': No Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }

                if ( mode == CMD_VERIFY )
                   return 0;
		
                bursts = config.rf_bursts[subcode];
		
		rfhcode = rev_low_nybble(hc2code(hc));

		mask = 1;
                for ( j = 0; j < 16; j++ ) {
                   if ( bitmap & mask ) {
		      rfword = (rfhcode << 12) | rf_unit_code[j] | rf_func_code[subcode];
                      display_rf_xmit(subcode, rfword, bursts);
		      if ( write_cm17a(rfword, bursts, rfmode) != 0 ) {
                         sprintf(errmsg,
			   "Command '%s': Unable to write to CM17A device.", label);
                         store_error_message(errmsg);
                         return 1;
                      }
		      rf_post_delay();
		   }
		   mask = mask << 1;
		}

		break;

	     case 9 :  /* RF Dim from full bright (turn Off before dimming) */
		
		if ( aflags & A_MINUS ) {
		   sprintf(errmsg,
                      "Command '%s': '-' prefix invalid for this command", label);
		   store_error_message(errmsg);
		   return 1;
		}
		if ( bitmap == 0 ) {
                   sprintf(errmsg,
		     "Command '%s': No Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
		}
		
	     case 4 :  /* RF Dim */
	     case 5 :  /* RF Bright */

		if ( (aflags & A_MINUS) )
		   bitmap = 0;

		sp = " ";
		if ( strchr(argv[2], '$') && mode == CMD_VERIFY )
		   bursts = 1;
		else  
		   bursts = strtol(argv[2], &sp, 10);
		
		if ( !strchr(" \t\n", *sp) || bursts < 1 || bursts > 255 ) {
                   sprintf(errmsg, "Command '%s': Invalid argument '%s'", label, argv[2]);
                   store_error_message(errmsg);
                   return 1;
                }

                if ( mode == CMD_VERIFY )
		   break;

		if ( subcode == 9 ) {
		   subcode = 4;
		   precode = 3;   /* Turn Off before dimming */
		}
		else
	           precode = 2;   /* Turn On before dimming */

		burstgroup = min(bursts, config.rf_bursts[subcode]);
		
		if ( rfmode == RF_FAST ) 
		   burstgroup = bursts;

		rfhcode = rev_low_nybble(hc2code(hc));
		
		if ( bitmap == 0 ) {
                   /* Transmit dims/brights only */
		   rfword = (rfhcode << 12) | rf_func_code[subcode];
                   display_rf_xmit(subcode, rfword, bursts);
		   while ( bursts > 0 ) {
                      if ( write_cm17a(rfword, min(bursts, burstgroup), rfmode) != 0 ) {
                         sprintf(errmsg,
                            "Command '%s': Unable to write to CM17A device.", label);
                         store_error_message(errmsg);
                         return 1;
                      }
                      rf_post_delay();
		      bursts -= burstgroup;
                   }
		}
		else {
                   /* Precede dims/brights with an On (or Off) command to set the address. */
		   mask = 1;
		   savebursts = bursts;
		   for ( j = 0; j < 16; j++ ) {
                      if ( bitmap & mask ) {
		         rfonword = (rfhcode << 12) | rf_unit_code[j] | rf_func_code[precode];
                         display_rf_xmit(precode, rfonword, config.rf_bursts[precode]);
			 rfword = (rfhcode << 12) | rf_func_code[subcode];			 
                         display_rf_xmit(subcode, rfword, bursts);
                         if ( write_cm17a(rfonword, config.rf_bursts[precode], RF_SLOW) != 0 ) {
                            sprintf(errmsg,
			      "Command '%s': Unable to write to CM17A device.", label);
                            store_error_message(errmsg);
                            return 1;
                         }
		         rf_post_delay();

			 while ( bursts > 0 ) {
                            if ( write_cm17a(rfword, min(bursts, burstgroup), rfmode) != 0 ) {
                               sprintf(errmsg,
				 "Command '%s': Unable to write to CM17A device.", label);
                               store_error_message(errmsg);
                               return 1;
                            }
		            rf_post_delay();
			    bursts -= burstgroup;
			 }
                      }
		      mask = mask << 1;
		      bursts = savebursts;
		   }
		}

		break;

	     case 7 : /* RF Arbitary 2-byte code */
		byte1 = byte2 = 0;
		sp = " ";
		if ( strchr(argv[1], '$') && mode == CMD_VERIFY )
		   byte1 = 0x10;
		else 
		   byte1 = strtol(argv[1], &sp, 16);
		if ( !strchr(" \t\n", *sp) || byte1 < 0 || byte1 > 0xff ) {
                   sprintf(errmsg,
		     "Command '%s': Invalid argument '%s'", label, argv[1]);
                   store_error_message(errmsg);
                   return 1;
                }

		sp = " ";
		if ( strchr(argv[2], '$') && mode == CMD_VERIFY )
		   byte2 = 0x10;
		else 
		   byte2 = strtol(argv[2], &sp, 16);
		if ( !strchr(" \t\n", *sp) || byte2 < 0 || byte2 > 0xff ) {
                   sprintf(errmsg,
		     "Command '%s': Invalid argument '%s'", label, argv[2]);
                   store_error_message(errmsg);
                   return 1;
                }
	
	        if ( strchr(argv[3], '$') && mode == CMD_VERIFY )
		   bursts = 1;
		else
                   bursts = strtol(argv[3], &sp, 10);
		
                if ( !strchr(" \t\n", *sp) || bursts < 1 || bursts > 255 ) {
                   sprintf(errmsg,
		        "Command '%s': Invalid <count> parameter", label);
	       	   store_error_message(errmsg);
	     	   return 1;
                }

                if ( mode == CMD_VERIFY )
	           return 0;

		burstgroup = min(bursts, config.rf_bursts[subcode]);
		if ( rfmode == RF_FAST )
		   burstgroup = bursts;

		rfword = byte1 << 8 | byte2;
                display_rf_xmit(subcode, rfword, bursts);

                while ( bursts > 0 ) {		
		   if ( write_cm17a(rfword, min(bursts, burstgroup), rfmode) != 0 ) {
                      sprintf(errmsg,
		        "Command '%s': Unable to write to CM17A device.", label);
                      store_error_message(errmsg);
                      return 1;
                   }
		   rf_farb_delay();
		   bursts -= burstgroup;
		}

                break;

             case 8 :  /* Reset CM17A */

		if ( mode == CMD_VERIFY )
		   return 0;

                if ( reset_cm17a() != 0 ) {
                   sprintf(errmsg,
		     "Command '%s': Unable to reset CM17A device.", label);
                   store_error_message(errmsg);
                   return 1;
                }

                break;
		
	     default :	   
		break;
	  }
          break;
#endif /* End of HASCM17A */	  
             
       default :
          sprintf(errmsg, "Command '%s': Not implemented", label);
          store_error_message(errmsg);
          return 1;
    }

    return 0;
}

/*---------------------------------------------------------------+
 | Generate the elements for an uploaded macro and pass back     |
 | through the argument list.  Return 0 if successful, otherwise |
 | write message to error handling routine and return 1.         | 
 +---------------------------------------------------------------*/
int macro_command( int argc, char *argv[],
                 int *nelem, int *nbytes, unsigned char *elements )
{

    int           j, k, icmd, nargs, radix, subcode;
    unsigned int  mask, aflags, bitmap, sendmap;
    unsigned long flags;
    char          *label, *sp;
    unsigned char *ep;
    char          hc;
    char          errmsg[128];
    int           dimval, dimcode, delay;
    unsigned char cmdcode, hcode, switchv, switchcode;
    int           minargs, maxargs; 
    unsigned int  xdata;
    static char   reorder[] = {6,14,2,10,1,9,5,13,7,15,3,11,0,8,4,12};

    clear_error_message();

    flags = 0;

    /* Handle a few special cases */

    if ( strcmp(argv[0], "function") == 0 ) {
       /* Special case: "function" command */
       if ( argc < 2 ) {
          store_error_message("Command 'function': Too few parameters");
          return 1;
       }
       argv++;
       argc--;
       flags |= F_NOA;
    }
    else if ( strcmp(argv[0], "turn") == 0 ) {
       /* Special case: "turn" command */
       if ( argc < 3 ) {
          store_error_message("Command 'turn': Too few parameters");
          return 1;
       }
       flags |= F_TRN;
       argv++;
       argc--;
       /* Swap first two parameters */
       sp = argv[0];
       argv[0] = argv[1];
       argv[1] = sp;
    }
    nargs = argc - 1;
      
    /* Identify the command */
    for ( icmd = 0; icmd < nx10cmds; icmd++ ) {
       if ( !strcmp(argv[0], x10command[icmd].label) )
          break;
    }
    if ( icmd == nx10cmds ) {
       sprintf(errmsg, "Invalid command '%s'", argv[0]);
       store_error_message(errmsg);
       return 1;
    }

    label    = x10command[icmd].label;
    minargs  = x10command[icmd].minargs;
    maxargs  = x10command[icmd].maxargs;
    flags   |= x10command[icmd].flags;
    cmdcode  = x10command[icmd].code;
    subcode  = x10command[icmd].subcode;
    radix    = (flags & F_ARB) ?  16 : 10;

    /* Make sure this command is available for uploaded macros */
    if ( flags & F_NMA || flags & F_HLP ) {
       sprintf(errmsg, "'%s' command is invalid for uploaded macros", label);
       store_error_message(errmsg);
       return 1;
    }

    if ( flags & F_NMAI ) {
       sprintf(errmsg, "Warning: '%s' command is ignored for uploaded macros", label);
       store_error_message(errmsg);
       *nelem = *nbytes = 0;
       return 0;
    }
    
    if ( nargs < minargs ) {
       sprintf(errmsg, "Command '%s': Too few arguments", label);
       store_error_message(errmsg);
       return 1;
    }
    if ( maxargs != -1 && nargs > maxargs ) {
       cmd_usage(stderr, icmd );
       sprintf(errmsg, "Command '%s': Too many arguments", label);
       store_error_message(errmsg);
       return 1;
    }

    /* Handle a few special cases */

    if ( flags & F_ARB ) {
       /* If arbitrary extended code, get subcode and adjust argument count */
       subcode = (int)strtol(argv[1], &sp, radix);
       argv++;
       argc--;
       nargs--;
       if ( !strchr(" \t\n", *sp) || subcode < 0 || subcode > 0xff ) {
          sprintf(errmsg, "Command '%s': Extended code function outside range 00-0xff", label);
          store_error_message(errmsg);
          return 1;
       }
    }

    /* Only commands with command code <= 6  or code == 26 are valid with "turn" */
    if ( flags & F_TRN && cmdcode > 6 && cmdcode != 26 ) {
       sprintf(errmsg, 
         "Command '%s' cannot be used with the 'turn' command", label); 
       store_error_message(errmsg);
       return 1;
    }

    /* Default HC|Unit */
    hc = '_';
    bitmap = 0;
    aflags = A_VALID;

    if ( !(flags & F_NUM) && nargs > 0 ) {
       /* parse it as an alias or HC|Units token. */
       aflags = parse_addr(argv[1], &hc, &bitmap);
    }

    if ( !(aflags & A_VALID) ) {
       sprintf(errmsg, "Command '%s': ", label);
       add_error_prefix(errmsg);
       return 1;
    }

    /* Make sure a positional parameter didn't slip through */
    if ( aflags & A_DUMMY ) {
       sprintf(errmsg,
          "Command '%s': Unresolved positional parameter '%s'",
             label, argv[1]);
       store_error_message(errmsg);
       return 1;
    }
       
    if ( !(aflags & A_HCODE) && !(flags & F_NUM) && minargs > 0 )  {
       sprintf(errmsg, "Command '%s': Housecode required", label);
       store_error_message(errmsg);
       return 1;
    }

    switchv = cmdcode;

    switch ( switchv ) {

       case 20 : /* Send address bytes only */
       case 21 : /* Function only */
       case 22 : /* Send arbitrary series of hex bytes */
       case 23 : /* Send (quoted) text message */
       case 25 : /* Pause */
       case 27 : /* Turn */
       case 28 : /* Temp_Req */
       case 99 : /* Legacy commands */
          /* We should never get here */
             sprintf(errmsg, "Internal error - Invalid macro command '%s'", label);
             store_error_message(errmsg);
             return 1;

       case  8 :  /* Hail */
       case  9 :  /* Hail Ack */
          if ( !(aflags & A_HCODE) ) {
             hc = config.housecode;
          }
       case  0 :  /* All Units Off  */
       case  1 :  /* All Lights On  */
       case  6 :  /* All Lights Off */
       case 12 :  /* Data Transfer */
          if ( bitmap != 0 && !(aflags & A_PLUS) && !(aflags & A_MINUS) && config.force_addr == NO) {
             sprintf(errmsg, "Command '%s': Unitcode ignored\n", label);
             store_error_message(errmsg);
             *errmsg = '\0';
             bitmap = 0;
          }
          if ( bitmap == 0 && (aflags & A_PLUS || config.force_addr == YES) )
             bitmap = 1;

       case  2 :  /* On */
       case  3 :  /* Off */
       case 13 :  /* Status On */
       case 14 :  /* Status Off */
          if ( aflags & A_MINUS || flags & F_NOA )
             bitmap = 0;
          else if ( bitmap == 0 && !(flags & F_ALL) ) {
             sprintf(errmsg, "Command '%s': No Unitcode specified", label);
             store_error_message(errmsg);
             return 1;
          }

          hcode = hc2code(hc);

          elements[0] = (hcode << 4) | cmdcode;
          elements[1] = (unsigned char)((bitmap & 0xff00) >> 8);
          elements[2] = (unsigned char)((bitmap & 0x00ff));
          *nelem = 1;
          *nbytes = 3;

          break;
                 
       case  4 :  /* Dim */
       case  5 :  /* Bright */
          dimval = (int)strtol(argv[2], &sp, 10); 
          if ( !strchr(" \t\n", *sp) ) {
	     sprintf(errmsg,
		"Command '%s': Invalid Dim/Bright parameter '%s'", label, argv[2]);
             store_error_message(errmsg);
	     return 1;
	  }
	  else if ( config.restrict_dims == YES && (dimval < 1 || dimval > 22) ) {
             sprintf(errmsg, 
                "Command '%s': Dim/Bright value outside range 1-22", label);
             store_error_message(errmsg);
	     return 1;
	  }
	  else if ( config.restrict_dims == NO && (dimval < 0 || dimval > 31) ) {
             sprintf(errmsg, 
                "Command '%s': Dim/Bright value outside range 0-31", label);
             store_error_message(errmsg);
             return 1;
          }

          if ( aflags & A_MINUS || flags & F_NOA )
             bitmap = 0;
          else if ( bitmap == 0 ) {
             sprintf(errmsg, "Command '%s': No Unitcode specified", label);
             store_error_message(errmsg);
             return 1;
          }

          hcode = hc2code(hc);

          elements[0] = (hcode << 4) | cmdcode;
          elements[1] = (unsigned char)((bitmap & 0xff00) >> 8);
          elements[2] = (unsigned char)((bitmap & 0x00ff));
          elements[3] = (unsigned char)(dimval | ((flags & F_BRI) ? 0x80 : 0));          
          *nelem = 1;
          *nbytes = 4;

          break;
    
       case 10 :  /* Old style preset 1 */
       case 11 :  /* Old style preset 2 */

          if ( nargs == 2 )
             /* Limited preset for macros */
             dimval = (int)strtol(argv[2], &sp, 10);
          else
             /* preset level only */
             dimval = (int)strtol(argv[1], &sp, 10);

          if ( !strchr(" \t\n", *sp) || dimval < 1 || dimval > 32 ) {
             sprintf(errmsg, "Command '%s': Level outside range 1-32", label);
             store_error_message(errmsg);
             return 1;
          }

          if ( (dimcode = dimval) > 16 ) {
             cmdcode = 11;
             dimcode -= 16;
          }
          dimcode = rev_low_nybble(dimcode - 1);

          if ( nargs == 2 && (hcode = hc2code(hc)) != dimcode  ) {
             sprintf(errmsg, "Command '%s': Preset level %d not supported for housecode %c",
                label, dimval, hc);
             store_error_message(errmsg);
             return 1;
          }

          if ( aflags & A_MINUS || flags & F_NOA )
             bitmap = 0;
          else if ( bitmap == 0 ) {
             sprintf(errmsg, "Command '%s': No Unitcode specified", label);
             store_error_message(errmsg);
             return 1;
          }

          elements[0] = (dimcode << 4) | cmdcode;
          elements[1] = (unsigned char)((bitmap & 0xff00) >> 8);
          elements[2] = (unsigned char)((bitmap & 0x00ff));
          *nelem = 1;
          *nbytes = 3;

          break;

       case 15 :  /* Status Req */
          sendmap = bitmap;
          if ( aflags & A_MINUS || flags & F_NOA ) {
             bitmap = 1;
             sendmap = 0;
          }
          else if ( bitmap == 0 ) {
             sprintf(errmsg, "Command '%s': No Unitcode specified", label);
             store_error_message(errmsg);
             return 1;
          }

          hcode = hc2code(hc);

          ep = elements;
          *nelem = 0;
          *nbytes = 0;
             
          /* Send each address/command pair individually to allow */
          /* associating a response to a particular unit code     */
          mask = 1;
          for ( k = 0; k < 16; k++ )  {
             if ( bitmap & mask ) {
                *ep++ = (hcode << 4) | cmdcode;
                *ep++ = (unsigned char)((bitmap & 0xff00) >> 8);
                *ep++ = (unsigned char)((bitmap & 0x00ff));
                *nelem += 1;
                *nbytes += 3;
             }
             mask = mask << 1;
          }         
          break;

       case 24 :  /* Send All_Units_Off to all Housecodes */
          bitmap = (config.force_addr == YES) ?  1 : 0;
          ep = elements;
	  for ( j = 0; j < 16; j++ ) {
	     hcode = reorder[j];    /* Send housecodes in alphabetic order */
             *ep++ = (hcode << 4);  /* cmdcode = 0 */
             *ep++ = (unsigned char)((bitmap & 0xff00) >> 8);
             *ep++ = (unsigned char)((bitmap & 0x00ff));
          }
          *nelem = 16;
          *nbytes = 48;

	  break;

       case 26 :  /* All Units On (same as 'on H1-16') */
          if ( bitmap != 0 && !(aflags & A_MINUS) ) {
             sprintf(errmsg, "Unitcode ignored for '%s' command.", label);
             store_error_message(errmsg);
          }
          bitmap = 0xFFFF;

          hcode = hc2code(hc);

          elements[0] = (hcode << 4) | 2 ;
          elements[1] = (unsigned char)((bitmap & 0xff00) >> 8);
          elements[2] = (unsigned char)((bitmap & 0x00ff));
          *nelem = 1;
          *nbytes = 3;
          
          break;

       case 32 :  /* Delay in minutes */
          delay = (int)strtol(argv[1], &sp, 10);
          if ( !strchr(" \t\n", *sp) ) {
             sprintf(errmsg, 
	        "Command '%s': Invalid character '%c' in delay value", label, *sp);
             store_error_message(errmsg);
             return 1;
          }
          if ( delay < 0 || delay > 240 ) {
             sprintf(errmsg, "Command '%s': value %d outside range 0-240 minutes",
		 label, delay);
             store_error_message(errmsg);
             return 1;
          }
	  	  
	  sprintf(errmsg, "Command '%s': Not yet implemented for macros", label);
	  store_error_message(errmsg);
	  return 1;
	  
       case  7 :  /* Extended Code commands */
          sendmap = (aflags & A_PLUS) ? bitmap : 0 ;
          if ( sendmap == 0 && config.force_addr == YES )
             sendmap = 0xFFFF;

          hcode = hc2code(hc);
          xdata = 0;

          if ( flags & F_ARB ) 
             switchcode = 0xff;
          else
             switchcode = subcode;

          switch ( switchcode ) {
             case 0x31 :  /* Extended Preset Dim */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': No Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }

                /* Get dim level */
                xdata = (int)strtol(argv[2], &sp, 10);  
                if ( !strchr(" \t\n", *sp) || xdata < 0 || xdata > 63 ) {
                   sprintf(errmsg, "Command '%s': Extended Preset level outside range 0-63", label);
                   store_error_message(errmsg);
                   return 1;
                }
                break;

             case 0x3b :  /* Extended Configure Auto Status_Ack Mode */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': No Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }

                /* Get mode */
                xdata = (int)strtol(argv[2], &sp, 10);  
                if ( !strchr(" \t\n", *sp) || xdata < 0 || xdata > 3 ) {
                   sprintf(errmsg, "Command '%s': Extended Config Mode outside range 0-3", label);
                   store_error_message(errmsg);
                   return 1;
                }
                break;

             case 0xfe :  /* Extended ON ( = Extended Preset Dim level 63) */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': No Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }
                subcode = 0x31;
                xdata = 63;
                break;                

             case 0xfd :  /* Extended OFF ( = Extended Preset Dim level 0) */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': No Unitcode specified", label);
                   store_error_message(errmsg);
                   return 1;
                }
                subcode = 0x31;
                xdata = 0;
                break;                

             case 0x33 :  /* Extended All On */
             case 0x34 :  /* Extended All Off */

                if ( bitmap != 0 && !(aflags & A_PLUS) && !(aflags & A_MINUS) && config.force_addr == NO) {
                   sprintf(errmsg, "Command '%s': Unitcode ignored", label);
                   store_error_message(errmsg);
                   *errmsg = '\0';
                   bitmap = 1;
                }
                else if ( bitmap == 0 )
                   bitmap = 1;

                if ( aflags & A_PLUS )
                   sendmap = bitmap;

             case 0x37 :  /* Extended Status Req */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': No Unitcode specified", label);
                   return 1;
                }
                xdata = 0;
                break;

             case 0xff :  /* Arbitrary Extended Function */
                if ( bitmap == 0 ) {
                   sprintf(errmsg, "Command '%s': No Unitcode specified", label);
                   return 1;
                }
                xdata = (int)strtol(argv[2], &sp, 16);
                if ( !strchr(" \t\n", *sp) || xdata < 0 || xdata > 0xff ) {
                   sprintf(errmsg, 
                      "Command '%s': Data '%s' outside range 0-FF", label, argv[2]);
                   store_error_message(errmsg);
                   return 1;
                }
                break ;

             default :
                sprintf(errmsg, "Command not implemented");
                store_error_message(errmsg);
                return 1;                
          }

          mask = 1;
          ep = elements;
          *nelem = 0;
          *nbytes = 0;
          for ( j = 0; j < 16; j++ ) {
             if ( bitmap & mask ) {
                *ep++ = (hcode << 4) | cmdcode;
                *ep++ = (unsigned char)(((sendmap & mask) & 0xff00) >> 8);
                *ep++ = (unsigned char)(((sendmap & mask) & 0x00ff));
                *ep++ = j;
                *ep++ = (unsigned char)(xdata & 0xff);
                *ep++ = (unsigned char)(subcode & 0xff);
                *nelem += 1;
                *nbytes += 6;
             }
             mask = mask << 1;
          }             
          break;

       default :
          sprintf(errmsg, "Command not implemented");
          store_error_message(errmsg);
          return 1;
    }

    return 0;
}

     
/*---------------------------------------------------------------+
 | Return the maximum number of positional parameters in a       |
 | scene, i.e., the highest integer 'nn' represented in a token  |
 | a token of the form "$nn".  Return -1 if the "nn" is missing, |
 | is less than 1, or is not an integer.                         |
 +---------------------------------------------------------------*/
int max_parms ( int tokc, char *tokv[] )
{
   int  j, parmno, count, maxparm = 0;
   int  bracket;
   char *sp, *tp;
   char errmsg[128];
   int  parmchk[1000];

   for ( j = 0; j < 32; j++ )
      parmchk[j] = 0;

   for ( j = 1; j < tokc; j++ ) {
      tp = tokv[j];      
      while ((sp = strchr(tp, '$')) != NULL ) {
         sp++;
         bracket = (*sp == '{') ? 1 : 0 ;
         parmno = (int)strtol(sp + bracket, &sp, 10);
         if ( (bracket && *sp != '}') || parmno < 1 ) {
            sprintf(errmsg, 
               "Command '%s': Invalid positional parameter in '%s'",
                   tokv[0], tokv[j]);
            store_error_message(errmsg);
            return -1;
         }
         if ( parmno > config.max_pparms ) {
            sprintf(errmsg, "Too many (%d) positional parameters - max %d",
               parmno, config.max_pparms);
            store_error_message(errmsg);
            return -1;
         }
         parmchk[parmno] = 1;
         maxparm = max(maxparm, parmno);
         if ( bracket )
            sp++;
         tp = sp;
      }
   }

   count = 0;
   for ( j = 1; j < 32; j++ )
      count += parmchk[j];

   /* Warn the user if all positional parameters within the */
   /* range 1 through maxparm arene't used in the scene.    */
   if ( count < maxparm ) {
      sprintf(errmsg, 
        "Warning: Not all positional parameters specified are used");
      store_error_message(errmsg);
   }
   return maxparm;
}
       
      
/*---------------------------------------------------------------+
 | Verify that the syntax of commands in a scene is correct.     |
 +---------------------------------------------------------------*/
int verify_scene ( char *body ) 
{
   int  j;
   int  cmdc, tokc, count;
   char **cmdv, **tokv;
   char *sp;

   sp = strdup(body);

   /* Separate scene into individual commands */
   tokenize(sp, ";", &cmdc, &cmdv);
   count = 0;
   for ( j = 0; j < cmdc; j++ ) {
      strtrim(cmdv[j]);
      if ( *cmdv[j] == '\0' )
         continue;
      count++;
      /* Separate commands into individual tokens */
      tokenize(cmdv[j], " \t", &tokc, &tokv);
      /* Verify the individual command */
      if ( direct_command(tokc, tokv, CMD_VERIFY) != 0 ) {
         free(tokv);
         free(cmdv);
         free(sp);
         return 1;
      }
      free(tokv);
   }
   free(cmdv);
   free(sp);
   if ( !count ) {
      store_error_message("Warning: Contains no commands.");
   }
   return 0;
}


/*---------------------------------------------------------------+
 | Replace the positional parameters in the members of list tokv |
 | with actual values.  It is presumed that the number of        |
 | members in 'actual' has already be verified to be sufficient. |
 | The new _members_ in tokvp must be freed after use.           |
 +---------------------------------------------------------------*/
void replace_dummy_parms ( int tokc, char ***tokvp, char *actual[] )
{
   int j, parmno, bracket;
   char *sp;
   char buffer1[128];
   char buffer2[128];

   for ( j = 0; j < tokc; j++ ) {
      strcpy(buffer1, (*tokvp)[j]);

      while ( (sp = strchr(buffer1, '$')) != NULL ) {   
         strncpy2(buffer2, buffer1, (sp - buffer1));
         sp++;
         bracket = (*sp == '{') ? 1 : 0 ;
         parmno = (int)strtol(sp + bracket, &sp, 10);
         strcat(buffer2, actual[parmno]);
         if ( bracket )
            sp++;
         strcat(buffer2, sp);
         strcpy(buffer1, buffer2);
      }
      if ( ((*tokvp)[j] = strdup(buffer1)) == NULL ) {
         fprintf(stderr, "Unable to allocate memory in replace_dummy_parms()\n");
         exit(1);
      }
   }

   return;
}

/*---------------------------------------------------------------+
 | Handler for direct commands and scenes entered from the       |
 | command line.                                                 |
 +---------------------------------------------------------------*/
int c_command( int argc, char *argv[] )
{
   int    j, k, m, cmdc, tokc, retcode;
   char   **cmdv, **tokv;
   char   *sp;
   SCENE  *scenep;

   scenep = config.scenep;

    /* Determine whether 2nd level command under "cmd". */
    /* If so, offset all arguments by 1.                */
    if ( strcmp(argv[1], "cmd") == 0 ) {
       argv++;
       argc--;
    }
    if ( argc < 2 )
       return 1;

    /* Special case: "help" command */
    if ( strcmp(argv[1], "help") == 0 ) {
       if ( argc < 3 ) 
	  command_help("");
       else
	  command_help(argv[2]);
       return 0;
    }
    
    /* Special case: "syn" (display synonyms) command */
    if ( strcmp(argv[1], "syn") == 0 ) {
       display_syn();
       return 0;
    }

    /* See if it's a user-defined scene */
    if ( (j = lookup_scene(scenep, argv[1])) >= 0 ) {

       if ( argc < (scenep[j].nparms + 2) ) {
          fprintf(stderr, 
             "(Config %02d: %s '%s'): Too few parameters - %d required.\n", 
                scenep[j].line_no, typename[scenep[j].type], scenep[j].label, 
                scenep[j].nparms);
          return 1;
       }
       else if ( argc > (scenep[j].nparms + 2) ) {
          fprintf(stderr, 
             "(Config %02d: %s '%s'): Too many parameters - only %d accepted.\n", 
                scenep[j].line_no, typename[scenep[j].type], scenep[j].label,
                scenep[j].nparms);
          return 1;
       }

       sp = strdup(scenep[j].body);
       /* Separate scene commands separated by ';' */
       tokenize(sp, ";", &cmdc, &cmdv);
       for ( k = 0; k < cmdc; k++ ) {
	  strtrim(cmdv[k]);
	  if ( *cmdv[k] == '\0' )
             continue;
          tokenize(cmdv[k], " ", &tokc, &tokv);

          /* Substitute for any positional parameters */
          /* This function allocates new memory for   */
          /* each parameter since the replacement may */
          /* be longer than the string in argv        */
          replace_dummy_parms(tokc, &tokv, argv + 1);

          retcode = direct_command(tokc, tokv, CMD_RUN);

          /* Free each member of tokv, then tokv itself */
          for ( m = 0; m < tokc; m++ )
             free(tokv[m]);
          free(tokv);

          if ( retcode != 0 ) {
             /* Display fatal message from error handler */
             fprintf(stderr, "(Config %02d: %s '%s'): %s\n", 
                scenep[j].line_no, typename[scenep[j].type], scenep[j].label,
                error_message());
             clear_error_message();
             free(cmdv);
             free(sp);
             return retcode;
          }
          else if ( *error_message() ) {
             /* Display warning from error handler */
             fprintf(stderr, "(Config %02d: %s '%s'): %s\n", 
                scenep[j].line_no, typename[scenep[j].type], scenep[j].label,
                error_message());
          }
          clear_error_message();             
       }
       free(cmdv);
       free(sp);

       return 0;
    }

    /* It's an individual command.                      */
    /* Reduce the argument list so that the X10 command */
    /* itself becomes argv[0]                           */

    retcode = direct_command(--argc, ++argv, CMD_RUN);
    if ( retcode != 0 ) {
       fprintf(stderr, "%s\n", error_message());
       clear_error_message();
       return retcode;
    }
    else if ( *error_message() ) {
       fprintf(stderr, "%s\n", error_message());
       clear_error_message();
    }
  
    return 0;
}
    

        
/*---------------------------------------------------------------------+
 | Display macros for final_report()                                   |
 | Argument 'type' is either USED or NOTUSED.                          |
 +---------------------------------------------------------------------*/
int display_macros ( FILE *fd, unsigned char type, 
                           MACRO *macrop, unsigned char *elementp )
{
   int           i, j, k, m, switchval, label_max;
   char          delim, modflag;
   unsigned char cmdcode, subcode, hcode, subunit, flags, bflag, dim;
   unsigned int  bmap;

   /* Find the maximum length of a macro label */
   label_max = 0;
   j = 0;
   while ( macrop[j].line_no > 0 ) {
      /* Skip null macro and those of the wrong type (used/unused) */
      if ( macrop[j].use != type || macrop[j].isnull ) {
         j++;
         continue;
      }
     
      /* Skip derived macros which aren't used */
      if ( type == NOTUSED && (macrop[j].refer & DERIVED_MACRO) ) {
         j++;
         continue;
      } 

      label_max = max(label_max, strlen(macrop[j].label) );
      j++;
   }

   if ( label_max == 0 )
      return 0;

   /* Now display the macros */
   j = 0;
   while ( macrop[j].line_no > 0 ) {
      if ( macrop[j].use != type || macrop[j].isnull ) {
         j++;
         continue;
      }

      /* Skip derived macros which aren't used */
      if ( type == NOTUSED && (macrop[j].refer & DERIVED_MACRO) ) {
         j++;
         continue;
      } 
      
      modflag = (macrop[j].modflag == UNMODIFIED) ? ' ' : '*' ;

      if ( type == USED && config.display_offset == YES )
         (void) fprintf( fd, "%03x macro  %-*s %3d %c", macrop[j].offset,
             label_max, macrop[j].label, macrop[j].delay, modflag);
      else
         (void) fprintf( fd, "macro  %-*s %3d %c", label_max,
             macrop[j].label, macrop[j].delay, modflag);


      delim = ' ';
      i = macrop[j].element ;
      for ( k = 0; k < macrop[j].nelem; k++ ) {
         cmdcode = elementp[i] & 0x0f;
         hcode = code2hc((elementp[i] & 0xf0) >> 4);
         bmap = elementp[i+1] << 8 | elementp[i+2];
         switchval = -1;

         bflag = 0;
         if ( cmdcode == 4 || cmdcode == 5 ) {
            /* Dim or Bright command                    */
            /* Check both command code and brighten bit */
            bflag = (elementp[i+3] & 0x80) ? F_BRI : 0;
            for ( m = 0; m < nx10cmds; m++ ) {
               if ( x10command[m].code == cmdcode && 
                    (x10command[m].flags & F_BRI) == bflag ) {
                  switchval = m;
                  break;
               }
            }
         }
         else if ( cmdcode == 10 || cmdcode == 11 ) {
            /* Preset command */
            if ( bmap == 0 ) {
               /* Check for preset_level only */
               for ( m = 0; m < nx10cmds; m++ ) {
                  if ( x10command[m].code == cmdcode && 
                       x10command[m].flags & F_NOA ) {
                     switchval = m;
                     break;
                  }
               }
            }
            else {
               /* Check for mpreset */
               for ( m = 0; m < nx10cmds; m++ ) {
                  if ( x10command[m].code == cmdcode && 
                       x10command[m].flags & F_MAC ) {
                     switchval = m;
                     break;
                  }
               }
            }
         }
         else {
            for ( m = 0; m < nx10cmds; m++ ) {
               if ( x10command[m].code == cmdcode ) {
                  switchval = m;
                  break;
               }
            }
         }

         if ( switchval < 0 || switchval == nx10cmds ) {
            (void) fprintf(stderr, "display_macros(): Internal error 1\n");

            break;
         }

         switch ( cmdcode ) {
            case 0 :  /* All Units Off */
            case 1 :  /* All Lights On */
            case 6 :  /* All Lights Off */             
            case 8 :  /* Hail Request */
            case 9 :  /* Hail Ack */
            case 12 : /* Data Transfer */               
	    case 13 : /* Status On */
	    case 14 : /* Status Off */
               /* Bitmap is normally zero for these commands */
               if ( bmap ) {
                  /* Display the '+' prefixed HC and unit */  
                  fprintf( fd, "%c %s +%c%s",
                      delim, x10command[switchval].label, hcode,
                      bmap2units(bmap) );
               }
               else {
                  /* Just the HC */
                  fprintf( fd, "%c %s %c",
                      delim, x10command[switchval].label, hcode);
               }

               i += x10command[switchval].length;
               break;

            case 2 :  /* On */
            case 3 :  /* Off */
            case 15 : /* Status Request */
               if ( bmap ) {
                  fprintf( fd, "%c %s %c%s",
                      delim, x10command[switchval].label, hcode,
                      bmap2units(bmap) );
               }
               else {
                  fprintf( fd, "%c %s -%c",
                      delim, x10command[switchval].label, hcode);
               }

               i += x10command[switchval].length;
               break;

            case 4 :  /* Dim or Dimb */
            case 5 :  /* Bright or Brightb */
               if ( bmap ) {
                  fprintf( fd, "%c %s %c%s %d",
                     delim, x10command[switchval].label, hcode,
                     bmap2units(bmap), elementp[i+3] & 0x1f );
               }
               else {
                  fprintf( fd, "%c %s -%c %d",
                     delim, x10command[switchval].label, hcode,
                     elementp[i+3] & 0x1f );
               }

               i += x10command[switchval].length;

               break;

            case 10 : /* Preset Dim (1) - mpreset or level only */
            case 11 : /* Preset Dim (2) - mpreset or level only */
               dim = rev_low_nybble((elementp[i] & 0xf0) >> 4) + 1;
               dim += (cmdcode == 11) ? 16 : 0;
               if ( bmap == 0 ) {
                  fprintf( fd, "%c %s %d", 
                     delim, x10command[switchval].label, dim);
               }
               else {
                  fprintf( fd, "%c %s %c%s %d",
                     delim, x10command[switchval].label,
                       hcode, bmap2units(bmap), dim );
               }

               i += x10command[switchval].length;
               break;

            case  7 : /* Extended Code */
               subcode = elementp[i+5];
               for ( m = 0; m < nx10cmds; m++ ) {
                  if ( x10command[m].code == cmdcode && 
                         x10command[m].subcode == subcode ) 
                  break;
               }

               /* If not standard command, locate entry for arbitrary code */
               if ( m >= nx10cmds ) {
                  for ( m = 0; m < nx10cmds; m++ ) {
                     if ( !(strcmp(x10command[m].label, XARBNAME)) )
                        break;
                  }
               }

               flags = x10command[m].flags;


               if ( flags & F_ARB )
                  fprintf( fd, "%c %s %02X ", delim, x10command[m].label, subcode);
               else
                  fprintf( fd, "%c %s ", delim, x10command[m].label); 

               /* Display the HC, prefixed with a '+' if bitmap is not 0 */
               if ( bmap )
                  fprintf(fd, "+%c", hcode);
               else
                  fprintf( fd, "%c", hcode);

               /* Display the unit, if any */
               if ( !(flags & F_ALL) || flags & F_FAD || flags & F_ARB ) {
                  fprintf( fd, "%d", code2unit(elementp[i+3] & 0x0f) );
		  subunit = (elementp[i+3] & 0xf0) >> 4;
		  if ( subunit > 0 )
		     fprintf( fd, "(%02x)", subunit );
	       }
               else if ( flags & F_ALL && bmap )
                  fprintf( fd, "%s", bmap2units(bmap) ); 

               /* Display the data byte if required */
               if ( flags & F_DED )
                  fprintf( fd, " %02d", elementp[i+4]);
               else if ( flags & F_ARB )
                  fprintf( fd, " %02X", elementp[i+4]);
 
               i += x10command[m].length;
               break;

            default :
               fprintf(stderr, "display_macros(): Internal error 2.\n");
               break;               
         }

         delim = ';';
      }
      (void) fprintf( fd, "\n" );
      j++ ;
   }

   return 1;
}

/*---------------------------------------------------------------------+
 | Display error message and return 1 if configured for a CM10A        |
 | interface.  Otherwise just return 0.                                |
 +---------------------------------------------------------------------*/
int invalidate_for_cm10a ( void )
{
   if ( config.device_type & DEV_CM10A ) {
      fprintf(stderr, "Command is invalid for CM10A interface.\n");
      return 1;
   }
   return 0;
}   

